# Copyright (c) 2019 NVIDIA Corporation
import copy
import importlib
import itertools
import json
import os
from collections import defaultdict
from contextlib import ExitStack
from pathlib import Path
from typing import List, Optional

import torch
import torch.distributed as dist
import torch.nn as nn
import torch.optim as optim
from torch.nn.parallel import DistributedDataParallel as DDP

from nemo import logging
from nemo.backends.pytorch.module_wrapper import TrainableNeuralModuleWrapper
from nemo.backends.pytorch.nm import DataLayerNM, TrainableNM
from nemo.backends.pytorch.optimizers import AdamW, Novograd, master_params
from nemo.core import DeploymentFormat, DeviceType, NeuralModule, NeuralModuleFactory, NmTensor
from nemo.core.actions import Actions, TrainingState, topological_sort_from_leaves
from nemo.core.callbacks import ActionCallback, NeMoCallback, SimpleLossLoggerCallback
from nemo.core.neural_factory import OperationMode, Optimization
from nemo.core.neural_types import AxisKind, NeuralType
from nemo.utils.app_state import AppState
from nemo.utils.decorators import deprecated
from nemo.utils.helpers import get_checkpoint_from_dir

# these imports will happen on as-needed basis
amp = None
LARC = None
FusedLAMB = None
FusedAdam = None
FusedNovoGrad = None

AmpOptimizations = {
    Optimization.mxprO0: "O0",
    Optimization.mxprO1: "O1",
    Optimization.mxprO2: "O2",
    Optimization.mxprO3: "O3",
}

_float_2_half_req = {
    Optimization.mxprO1,
    Optimization.mxprO2,
    Optimization.mxprO3,
}


class PtActions(Actions):
    def __init__(
        self, local_rank=None, global_rank=None, tb_writer=None, optimization_level=Optimization.mxprO0,
    ):
        need_apex = local_rank is not None or optimization_level != Optimization.mxprO0
        if need_apex:
            try:
                apex = importlib.import_module('apex')
                if optimization_level != Optimization.mxprO0:
                    global amp
                    amp = importlib.import_module('apex.amp')
                if local_rank is not None:
                    global LARC
                    global FusedLAMB
                    global FusedAdam
                    global FusedNovoGrad
                    parallel = importlib.import_module('apex.parallel')
                    apex_optimizer = importlib.import_module('apex.optimizers')
                    LARC = parallel.LARC
                    FusedLAMB = apex_optimizer.FusedLAMB
                    FusedAdam = apex_optimizer.FusedAdam
                    FusedNovoGrad = apex_optimizer.FusedNovoGrad

            except ImportError:
                raise ImportError(
                    "NVIDIA Apex is necessary for distributed training and"
                    "mixed precision training. It only works on GPUs."
                    "Please install Apex from "
                    "https://www.github.com/nvidia/apex"
                )

        super(PtActions, self).__init__(
            local_rank=local_rank, global_rank=global_rank, optimization_level=optimization_level,
        )

        self._step = 0
        self._epoch = 0
        self._optimizers = []
        self.tb_writer = tb_writer
        self.cache = None
        self.amp_initialized = False
        self.ddp_initialized = False
        self.ddp_module_dict = {}
        self._train_called = False

    @property
    def step(self):
        return self._step

    @step.setter
    def step(self, step):
        self._step = step

    @property
    def epoch(self):
        return self._epoch

    @epoch.setter
    def epoch(self, epoch):
        self._epoch = epoch

    @property
    @deprecated(version="0.12", explanation="epoch_num has been deprecated in favour of epoch.")
    def epoch_num(self):
        return self._epoch

    @epoch_num.setter
    @deprecated(version="0.12", explanation="epoch_num has been deprecated in favour of epoch.")
    def epoch_num(self):
        return self._epoch

    @property
    def optimizers(self):
        return self._optimizers

    def __get_top_sorted_modules_and_dataloader(self, hook: List[NmTensor]):
        """A function that accepts a list of NmTensors that need to be computed and constructs a call DAG that starts
        from a datalayerNM and can be used to compute the NmTensors.

        args:
            leaf_nmtensors (List[NmTensors]): The tensors to be computed

        returns:
            top_sorted_modules: the callchain DAG
            tdataset: the datalayer at the top of the callchain
        """
        top_sorted_modules = topological_sort_from_leaves(hook)

        if not isinstance(top_sorted_modules[0][0], DataLayerNM):
            raise ValueError("The first module in your DAG was not a DataLayer NeuralModule.")

        tdataset = top_sorted_modules[0][0].dataset

        for m in top_sorted_modules:
            if m[0].factory is None and self._local_rank is not None:
                raise ValueError(
                    "Neural module {0} was created without NeuralModuleFactory, but you are trying to run in "
                    "distributed mode. Please instantiate NeuralModuleFactory first and pass its instance as "
                    "`factory` parameter to all your Neural Module objects.".format(str(m[0]))
                )

        return top_sorted_modules, tdataset

    def create_optimizer(self, optimizer, things_to_optimize, optimizer_params=None):
        """
        Wrapper function around __setup_optimizer()

        Args:
            optimizer : A instantiated PyTorch optimizer or string. For
                currently supported strings, see __setup_optimizer().
            things_to_optimize (list): Must be a list of Neural Modules and/or
                parameters. If a Neural Module is passed, all trainable
                parameters are extracted and passed to the optimizer.
            optimizer_params (dict): Optional parameters dictionary.

        Returns:
            Optimizer
        """

        optimizer_instance = None
        optimizer_class = None
        if isinstance(optimizer, str):
            optimizer_class = optimizer
        elif isinstance(optimizer, torch.optim.Optimizer):
            optimizer_instance = optimizer
        else:
            raise ValueError("`optimizer` must be a string or an instance of torch.optim.Optimizer")

        modules_to_optimize = []
        tensors_to_optimize = []
        if not isinstance(things_to_optimize, list):
            things_to_optimize = [things_to_optimize]
        for thing in things_to_optimize:
            if isinstance(thing, NeuralModule):
                modules_to_optimize.append(thing)
            elif isinstance(thing, NmTensor):
                tensors_to_optimize.append(thing)
            else:
                raise ValueError(
                    "{} passed to create_optimizer() was neither a neural module nor a neural module tensor"
                )

        if tensors_to_optimize:
            call_chain, _ = self.__get_top_sorted_modules_and_dataloader(tensors_to_optimize)

            for module in call_chain:
                if module[0] not in modules_to_optimize:
                    modules_to_optimize.append(module[0])

        # Extract trainable weights which will be optimized
        params_list = [p.parameters() for p in modules_to_optimize if isinstance(p, TrainableNM) or p.is_trainable()]
        params_to_optimize = itertools.chain(*params_list)

        if optimizer_params is None:
            optimizer_params = {}
        # Init amp
        optimizer = self.__setup_optimizer(
            optimizer_instance=optimizer_instance,
            optimizer_class=optimizer_class,
            optimization_params=optimizer_params,
            params_to_optimize=params_to_optimize,
        )

        self._optimizers.append(optimizer)
        return optimizer

    @staticmethod
    def __setup_optimizer(
        optimizer_instance, optimizer_class, optimization_params, params_to_optimize,
    ):

        if optimizer_instance is None:
            # Setup optimizer instance, by default it is SGD
            lr = optimization_params["lr"]
            if optimizer_class.lower() == "sgd":
                optimizer = optim.SGD(
                    params_to_optimize,
                    lr=lr,
                    momentum=optimization_params.get("momentum", 0.9),
                    weight_decay=optimization_params.get("weight_decay", 0.0),
                )
            elif optimizer_class.lower() == "adam":
                optimizer = optim.Adam(
                    params=params_to_optimize,
                    lr=lr,
                    betas=optimization_params.get("betas", (0.9, 0.999)),
                    eps=optimization_params.get("eps", 1e-8),
                    weight_decay=optimization_params.get("weight_decay", 0),
                    amsgrad=optimization_params.get("amsgrad", False),
                )
            elif optimizer_class.lower() == "fused_adam":
                if not FusedAdam:
                    raise ValueError("FusedAdam works only with torch DDP.")
                optimizer = FusedAdam(  # Note: amsgrad = True is not supported.
                    params=params_to_optimize,
                    lr=lr,
                    betas=optimization_params.get("betas", (0.9, 0.999)),
                    eps=optimization_params.get("eps", 1e-8),
                    weight_decay=optimization_params.get("weight_decay", 0.0),
                )
            elif optimizer_class.lower() == "adam_w":
                optimizer = AdamW(
                    params=params_to_optimize,
                    lr=lr,
                    betas=optimization_params.get("betas", (0.9, 0.999)),
                    eps=optimization_params.get("eps", 1e-8),
                    weight_decay=optimization_params.get("weight_decay", 0.0),
                    amsgrad=optimization_params.get("amsgrad", False),
                )
            elif optimizer_class.lower() == "novograd":
                optimizer = Novograd(
                    params_to_optimize,
                    lr=lr,
                    weight_decay=optimization_params.get("weight_decay", 0.0),
                    luc=optimization_params.get("luc", False),
                    luc_trust=optimization_params.get("luc_eta", 1e-3),
                    betas=optimization_params.get("betas", (0.95, 0.25)),
                )
            elif optimizer_class.lower() == "fused_novograd":
                if not FusedNovoGrad:
                    raise ValueError("FusedNovoGrad works only with torch DDP.")
                optimizer = FusedNovoGrad(
                    params_to_optimize,
                    lr=lr,
                    weight_decay=optimization_params.get("weight_decay", 0.0),
                    reg_inside_moment=True,
                    grad_averaging=False,
                    betas=optimization_params.get("betas", (0.95, 0.25)),
                )
            elif optimizer_class.lower() == "fused_lamb":
                if not FusedLAMB:
                    raise ValueError("FusedLAMB works only with torch DDP.")
                optimizer = FusedLAMB(params_to_optimize, lr=lr,)
            else:
                raise ValueError("Unknown optimizer class: {0}".format(optimizer_class))

            if optimization_params.get("larc", False):
                if not LARC:
                    raise ValueError("LARC works only with torch DDP.")
                logging.info("Enabling larc")
                optimizer = LARC(optimizer, trust_coefficient=optimization_params.get("larc_eta", 2e-2),)
        else:
            logging.info("Optimizer instance: {0} is provided.")
            if optimizer_class is not None and optimizer_class != "":
                logging.warning("Ignoring `optimizer_class` parameter because `optimizer_instance` is provided")
            if optimization_params is not None and optimization_params != {}:
                logging.warning(
                    "Ignoring `optimization_params` parameter for optimizer because `optimizer_instance` is provided"
                )
            optimizer = optimizer_instance
        return optimizer

    def __initialize_amp(
        self, optimizer, optim_level, amp_max_loss_scale=2.0 ** 24, amp_min_loss_scale=1.0,
    ):
        if optim_level not in AmpOptimizations:
            raise ValueError(f"__initialize_amp() was called with unknown optim_level={optim_level}")
        # in this case, nothing to do here
        if optim_level == Optimization.mxprO0:
            return optimizer

        if len(AppState().modules) < 1:
            raise ValueError("There were no modules to initialize")
        pt_modules = []
        for module in AppState().modules:
            if isinstance(module, nn.Module):
                pt_modules.append(module)
            elif isinstance(module, TrainableNeuralModuleWrapper):
                pt_modules.append(module._pt_module)

        _, optimizer = amp.initialize(
            max_loss_scale=amp_max_loss_scale,
            min_loss_scale=amp_min_loss_scale,
            models=pt_modules,
            optimizers=optimizer,
            opt_level=AmpOptimizations[optim_level],
        )
        self.amp_initialized = True
        return optimizer

    def nm_graph_forward_pass(self, callchain, registered_tensors):
        self.__nm_graph_forward_pass(callchain, registered_tensors)

    def __nm_graph_forward_pass(
        self, call_chain, registered_tensors, mode=OperationMode.training, use_cache=False,
    ):
        for ind in range(1, len(call_chain)):
            if use_cache:
                in_cache = True
                for tensor in call_chain[ind][2].values():
                    if tensor is None:
                        # NM has an output tensor that is not used in the
                        # current call chain, so we don't care if it's not in
                        # cache
                        continue
                    if tensor.unique_name not in registered_tensors:
                        in_cache = False
                if in_cache:
                    continue
            call_args = call_chain[ind][1]
            m_id = call_chain[ind][0].unique_instance_id
            pmodule = self.ddp_module_dict[m_id] if self.ddp_initialized else call_chain[ind][0]

            if mode == OperationMode.training:
                # if module.is_trainable():
                if isinstance(pmodule, nn.Module):
                    pmodule.train()
            elif mode == OperationMode.evaluation:
                # if module.is_trainable():
                if isinstance(pmodule, nn.Module):
                    pmodule.eval()
            else:
                raise ValueError("Unknown OperationMode")
            # prepare call signature for `module`
            call_set = {}
            for tensor_name, nmtensor in call_args.items():
                key = nmtensor.unique_name
                call_set[tensor_name] = registered_tensors[key]
            new_tensors = pmodule(force_pt=True, **call_set)

            if not isinstance(new_tensors, List):
                if not isinstance(new_tensors, tuple):
                    new_tensors = [new_tensors]
                else:
                    new_tensors = list(new_tensors)
            for t_tensor, nm_tensor in zip(new_tensors, call_chain[ind][2].values()):
                if nm_tensor is None:
                    continue
                t_name = nm_tensor.unique_name
                if t_name not in registered_tensors or registered_tensors[t_name] is None:
                    registered_tensors[t_name] = t_tensor
                else:
                    raise ValueError(f"A NMTensor was produced twice in the same DAG. {t_name}")

    @staticmethod
    def pad_tensor(t: torch.Tensor, target_size: torch.Size):
        padded_shape = target_size.cpu().data.numpy().tolist()
        padded_t = torch.zeros(padded_shape).cuda().type_as(t)
        t_size = t.size()
        if len(t_size) == 0:
            padded_t = t
        elif len(t_size) == 1:
            padded_t[: t_size[0]] = t
        elif len(t_size) == 2:
            padded_t[: t_size[0], : t_size[1]] = t
        elif len(t_size) == 3:
            padded_t[: t_size[0], : t_size[1], : t_size[2]] = t
        elif len(t_size) == 4:
            padded_t[: t_size[0], : t_size[1], : t_size[2], : t_size[3]] = t
        else:
            raise NotImplementedError
        return padded_t

    @staticmethod
    def depad_tensor(t: torch.Tensor, original_size: torch.Size):
        t_size = original_size
        if len(t_size) == 0:
            depadded_t = t
        elif len(t_size) == 1:
            depadded_t = t[: t_size[0]]
        elif len(t_size) == 2:
            depadded_t = t[: t_size[0], : t_size[1]]
        elif len(t_size) == 3:
            depadded_t = t[: t_size[0], : t_size[1], : t_size[2]]
        elif len(t_size) == 4:
            depadded_t = t[: t_size[0], : t_size[1], : t_size[2], : t_size[3]]
        else:
            raise NotImplementedError
        return depadded_t

    def _eval(self, tensors_2_evaluate, callback, step, verbose=False):
        """
        Evaluation process.
        WARNING THIS function assumes that all tensors_2_evaluate are based
        on a single datalayer
        Args:
          tensors_2_evaluate: list of NmTensors to evaluate
          callback: instance of EvaluatorCallback
          step: current training step, used for logging

        Returns:
          None
        """
        with torch.no_grad():
            # each call chain corresponds to a tensor in tensors_2_evaluate
            call_chain, _ = self.__get_top_sorted_modules_and_dataloader(hook=tensors_2_evaluate)
            # "Retrieve" data layer from call chain.
            dl_nm = call_chain[0][0]

            # Prepare eval_dataloader
            # For distributed training it should have disjoint subsets of
            # all data on every worker
            is_distributed = False
            world_size = None
            if dl_nm.placement == DeviceType.AllGpu:
                assert dist.is_initialized()
                is_distributed = True
                world_size = torch.distributed.get_world_size()

                if dl_nm.dataset is not None:
                    sampler = None
                    if not isinstance(dl_nm.dataset, torch.utils.data.IterableDataset):
                        sampler = torch.utils.data.distributed.DistributedSampler(
                            dataset=dl_nm.dataset, shuffle=dl_nm.shuffle
                        )
                    dataloader_params = {
                        'dataset': dl_nm.dataset,
                        'sampler': sampler,
                        'num_workers': dl_nm.num_workers,
                        'batch_size': dl_nm.batch_size,
                        'shuffle': False,
                        'pin_memory': dl_nm.pin_memory,
                    }
                    if hasattr(dl_nm, 'collate_fn'):
                        dataloader_params['collate_fn'] = dl_nm.collate_fn
                    eval_dataloader = torch.utils.data.DataLoader(**dataloader_params)
                else:
                    eval_dataloader = dl_nm.data_iterator

                if hasattr(eval_dataloader, 'sampler'):
                    eval_dataloader.sampler.set_epoch(0)
            else:  # Not distributed
                if dl_nm.dataset is not None:
                    # Todo: remove local_parameters
                    dataloader_params = {
                        'dataset': dl_nm.dataset,
                        'sampler': None,  # not distributed sampler
                        'num_workers': dl_nm.num_workers,
                        'batch_size': dl_nm.batch_size,
                        'shuffle': dl_nm.shuffle,
                        'pin_memory': dl_nm.pin_memory,
                    }
                    if hasattr(dl_nm, 'collate_fn'):
                        dataloader_params['collate_fn'] = dl_nm.collate_fn
                    eval_dataloader = torch.utils.data.DataLoader(**dataloader_params)
                else:
                    eval_dataloader = dl_nm.data_iterator
            # after this eval_dataloader is ready to be used
            # reset global_var_dict - results of evaluation will be stored
            # there

            callback.clear_global_var_dict()
            dl_device = dl_nm._device

            # Evaluation mini-batch for loop
            num_batches = None
            if hasattr(eval_dataloader, "__len__"):
                num_batches = len(eval_dataloader)
            for epoch_i, data in enumerate(eval_dataloader, 0):
                if (
                    verbose
                    and num_batches is not None
                    and (num_batches < 10 or (epoch_i % int(num_batches / 10) == 0))
                ):
                    logging.info(f"Evaluating batch {epoch_i} out of {num_batches}")
                tensors = []
                if isinstance(data, torch.Tensor):
                    data = (data,)
                for d in data:
                    if isinstance(d, torch.Tensor):
                        tensors.append(d.to(dl_device))
                    else:
                        tensors.append(d)

                registered_e_tensors = {
                    t.unique_name: d for t, d in zip(call_chain[0][2].values(), tensors) if t is not None
                }
                self.__nm_graph_forward_pass(
                    call_chain=call_chain, registered_tensors=registered_e_tensors, mode=OperationMode.evaluation,
                )

                if not is_distributed or self.global_rank == 0:
                    values_dict = {}
                # If distributed. For the outer loop, we need to ensure that
                # all processes loop through the elements in the same order
                for t2e in tensors_2_evaluate:
                    key = t2e.unique_name
                    if key not in registered_e_tensors.keys():
                        logging.info("WARNING: Tensor {} was not found during eval".format(key))
                        continue
                    if is_distributed:
                        # where we will all_gather results from all workers
                        tensors_list = []
                        # where we will all_gather tensor sizes
                        tensor_on_worker = registered_e_tensors[key]

                        if not isinstance(tensor_on_worker, torch.Tensor):  # For string and other.
                            if self.global_rank == 0:
                                values_dict[key] = [tensor_on_worker] + ([None] * (world_size - 1))
                            continue

                        # https://github.com/pytorch/pytorch/issues/24137
                        is_bool = False
                        if tensor_on_worker.dtype == torch.bool:
                            is_bool = True
                            tensor_on_worker = tensor_on_worker.to(dtype=torch.long)

                        if tensor_on_worker.shape != torch.Size([]):
                            tensor_on_worker_size_as_tensor = torch.tensor(tensor_on_worker.shape).cuda()
                            sizes = []
                            for ind in range(world_size):
                                sizes.append(torch.empty_like(tensor_on_worker_size_as_tensor))
                            dist.all_gather(sizes, tensor_on_worker_size_as_tensor)
                            mx_dim, _ = torch.max(torch.stack(sizes), dim=0)
                        else:  # this is a singleton. For example, loss value
                            sizes = [torch.Size([])] * world_size
                            mx_dim = None
                        for ind in range(world_size):
                            # we have to use max shape for all_gather
                            if mx_dim is None:  # singletons
                                tensors_list.append(torch.tensor(2).cuda().type_as(tensor_on_worker))
                            else:  # non-singletons
                                tensors_list.append(
                                    torch.empty(mx_dim.cpu().data.numpy().tolist()).cuda().type_as(tensor_on_worker)
                                )

                        if mx_dim is not None:
                            t_to_send = self.pad_tensor(tensor_on_worker, mx_dim)
                        else:
                            t_to_send = tensor_on_worker
                        dist.all_gather(tensors_list, t_to_send)
                        tensors_list = [self.depad_tensor(t, size) for t, size in zip(tensors_list, sizes)]
                        if self.global_rank == 0:
                            values_dict["IS_FROM_DIST_EVAL"] = True
                            if is_bool:
                                tensors_list = [t.to(dtype=torch.bool) for t in tensors_list]
                            values_dict[key] = tensors_list
                    else:  # NON-DISTRIBUTED TRAINING
                        values_dict["IS_FROM_DIST_EVAL"] = False
                        values_dict[key] = [registered_e_tensors[key]]
                if callback.user_iter_callback and (self.global_rank is None or self.global_rank == 0):
                    # values_dict will contain results from all workers
                    callback.user_iter_callback(values_dict, callback._global_var_dict)

            # final aggregation (over minibatches) and logging of results
            # should happend only on one worker
            if callback.user_done_callback and (self.global_rank is None or self.global_rank == 0):
                vals_to_log = callback.user_done_callback(callback._global_var_dict)
                # log results to Tensorboard or Weights & Biases
                if vals_to_log is not None:
                    if hasattr(callback, 'swriter') and callback.swriter is not None:
                        if hasattr(callback, 'tb_writer_func') and callback.tb_writer_func is not None:
                            callback.tb_writer_func(callback.swriter, vals_to_log, step)
                        else:
                            for key, val in vals_to_log.items():
                                callback.swriter.add_scalar(key, val, step)
                    if hasattr(callback, 'wandb_log'):
                        callback.wandb_log(vals_to_log)

    def _infer(
        self, tensors_to_return, verbose=False, cache=False, use_cache=False, offload_to_cpu=True,
    ):
        """
        Does the same as _eval() just with tensors instead of eval callback.
        """
        # Checking that cache is used properly
        if cache and use_cache:
            raise ValueError(
                "cache and use_cache were both set. However cache must first be created prior to using it."
            )
        if cache:
            if self.cache is not None:
                raise ValueError("cache was set but was not empty")
            self.cache = []
        if use_cache:
            if not self.cache:
                raise ValueError("use_cache was set, but cache was empty")

        with torch.no_grad():
            # each call chain corresponds to a tensor in tensors_2_evaluate
            dl_nm = None
            call_chain, _ = self.__get_top_sorted_modules_and_dataloader(hook=tensors_to_return)
            dl_nm = call_chain[0][0]

            # Prepare eval_dataloader
            # For distributed training it should have disjoint subsets of
            # all data on every worker
            is_distributed = False
            world_size = None
            if dl_nm.placement == DeviceType.AllGpu:
                if self.cache or use_cache:
                    raise NotImplementedError("Caching is not available for distributed training.")
                assert dist.is_initialized()
                is_distributed = True
                world_size = torch.distributed.get_world_size()
                if dl_nm.dataset is not None:
                    sampler = None
                    if not isinstance(dl_nm.dataset, torch.utils.data.IterableDataset):
                        sampler = torch.utils.data.distributed.DistributedSampler(
                            dataset=dl_nm.dataset, shuffle=dl_nm.shuffle
                        )
                    dataloader_params = {
                        'dataset': dl_nm.dataset,
                        'sampler': sampler,
                        'num_workers': dl_nm.num_workers,
                        'batch_size': dl_nm.batch_size,
                        'shuffle': False,
                        'pin_memory': dl_nm.pin_memory,
                    }
                    if hasattr(dl_nm, 'collate_fn'):
                        dataloader_params['collate_fn'] = dl_nm.collate_fn
                    eval_dataloader = torch.utils.data.DataLoader(**dataloader_params)
                else:
                    eval_dataloader = dl_nm.data_iterator
                eval_dataloader.sampler.set_epoch(0)
            elif not use_cache:  # Not distributed and not using cache
                # Dataloaders are only used if use_cache is False
                # When caching, the DAG must cache all outputs from dataloader
                if dl_nm.dataset is not None:
                    # Todo: remove local_parameters
                    dataloader_params = {
                        'dataset': dl_nm.dataset,
                        'sampler': None,  # not distributed sampler
                        'num_workers': dl_nm.num_workers,
                        'batch_size': dl_nm.batch_size,
                        'shuffle': dl_nm.shuffle,
                        'pin_memory': dl_nm.pin_memory,
                    }
                    if hasattr(dl_nm, 'collate_fn'):
                        dataloader_params['collate_fn'] = dl_nm.collate_fn
                    eval_dataloader = torch.utils.data.DataLoader(**dataloader_params)
                else:
                    eval_dataloader = dl_nm.data_iterator
            # after this eval_dataloader is ready to be used
            # reset global_var_dict - results of evaluation will be stored
            # there

            if not is_distributed or self.global_rank == 0:
                values_dict = {}
                for t in tensors_to_return:
                    values_dict[t.unique_name] = []
            dl_device = dl_nm._device

            # Evaluation mini-batch for loop
            if use_cache:
                num_batches = len(self.cache)
                loop_iterator = self.cache
            else:
                num_batches = len(eval_dataloader)
                loop_iterator = eval_dataloader

            for epoch_i, data in enumerate(loop_iterator, 0):
                if verbose and (num_batches < 10 or (epoch_i % int(num_batches / 10) == 0)):
                    logging.info(f"Evaluating batch {epoch_i} out of {num_batches}")
                tensors = []
                if use_cache:
                    registered_e_tensors = data
                    # delete tensors_to_return
                    for t in tensors_to_return:
                        if t.unique_name in registered_e_tensors:
                            del registered_e_tensors[t.unique_name]
                    # Need to check for device type mismatch
                    for t in registered_e_tensors:
                        registered_e_tensors[t].to(dl_device)
                else:
                    if isinstance(data, torch.Tensor):
                        data = (data,)
                    for d in data:
                        if isinstance(d, torch.Tensor):
                            tensors.append(d.to(dl_device))
                        else:
                            tensors.append(d)

                    registered_e_tensors = {
                        t.unique_name: d for t, d in zip(call_chain[0][2].values(), tensors) if t is not None
                    }
                self.__nm_graph_forward_pass(
                    call_chain=call_chain,
                    registered_tensors=registered_e_tensors,
                    mode=OperationMode.evaluation,
                    use_cache=use_cache,
                )

                if cache:
                    self.append_to_cache(registered_e_tensors, offload_to_cpu)

                # If distributed. For the outer loop, we need to ensure that
                # all processes loop through the elements in the same order
                for t2e in tensors_to_return:
                    key = t2e.unique_name
                    if key not in registered_e_tensors.keys():
                        logging.info("WARNING: Tensor {} was not found during eval".format(key))
                        continue
                    if is_distributed:
                        # where we will all_gather results from all workers
                        tensors_list = []
                        # where we will all_gather tensor sizes
                        tensor_on_worker = registered_e_tensors[key]
                        if tensor_on_worker.shape != torch.Size([]):
                            tensor_on_worker_size_as_tensor = torch.tensor(tensor_on_worker.shape).cuda()
                            sizes = []
                            for ind in range(world_size):
                                sizes.append(torch.empty_like(tensor_on_worker_size_as_tensor))
                            dist.all_gather(sizes, tensor_on_worker_size_as_tensor)
                            mx_dim, _ = torch.max(torch.stack(sizes), dim=0)
                        else:  # this is a singleton. For example, loss value
                            sizes = [torch.Size([])] * world_size
                            mx_dim = None
                        for ind in range(world_size):
                            # we have to use max shape for all_gather
                            if mx_dim is None:  # singletons
                                tensors_list.append(torch.tensor(2).cuda().type_as(tensor_on_worker))
                            else:  # non-singletons
                                tensors_list.append(
                                    torch.empty(mx_dim.cpu().data.numpy().tolist()).cuda().type_as(tensor_on_worker)
                                )

                        if mx_dim is not None:
                            t_to_send = self.pad_tensor(tensor_on_worker, mx_dim)
                        else:
                            t_to_send = tensor_on_worker
                        dist.all_gather(tensors_list, t_to_send)
                        tensors_list = [self.depad_tensor(t, size) for t, size in zip(tensors_list, sizes)]
                        if offload_to_cpu:
                            tensors_list = [t.cpu() for t in tensors_list]
                        if self.global_rank == 0:
                            values_dict[key] += tensors_list
                    else:  # NON-DISTRIBUTED TRAINING
                        tensor = registered_e_tensors[key]
                        if offload_to_cpu and isinstance(tensor, torch.Tensor):
                            tensor = tensor.cpu()
                        values_dict[key] += [tensor]

            if not is_distributed or self.global_rank == 0:
                inferred_tensors = []
                for t in tensors_to_return:
                    inferred_tensors.append(values_dict[t.unique_name])
                return inferred_tensors

            # For all other ranks
            return None

    def append_to_cache(self, registered_tensors: dict, offload_to_cpu):
        """Simpler helper function to add results of __nm_graph_forward_pass to
        current cache.
        """
        if offload_to_cpu:
            for t in registered_tensors:
                registered_tensors[t] = registered_tensors[t].cpu()
        self.cache.append(registered_tensors)

    def clear_cache(self):
        """ Simple helpful function to clear cache by setting self.cache to
        None
        """
        self.cache = None

    def save_state_to(self, path: str):
        """
        Saves current state such as step, epoch and optimizer parameters
        Args:
          path:

        Returns:

        """
        state = {
            "step": self.step,
            "epoch": self.epoch,
            "optimizer_state": [opt.state_dict() for opt in self._optimizers],
        }
        torch.save(state, path)

    def restore_state_from(self, path: str):
        """
        Restores state such as step, epoch and optimizer parameters
        Args:
          path:

        Returns:

        """
        if os.path.isfile(path):
            # map_location could be cuda:<device_id> but cpu seems to be more
            # general since we are also saving step and epoch
            # load_state_dict should move the variables to the relevant device
            checkpoint = torch.load(path, map_location="cpu")
            self.step = checkpoint["step"]
            self.epoch = checkpoint["epoch"]
            if checkpoint["optimizer_state"]:
                for opt, opt_chkpt in zip(self._optimizers, checkpoint["optimizer_state"]):
                    opt.load_state_dict(opt_chkpt)
        else:
            raise FileNotFoundError("Could not find checkpoint file: {0}".format(path))

    @staticmethod
    def _check_all_tensors(list_of_tensors):
        """Method that checks if the passed list contains all NmTensors
        """
        if not isinstance(list_of_tensors, list):
            return False
        for tensor in list_of_tensors:
            if not isinstance(tensor, NmTensor):
                return False
        return True

    @staticmethod
    def _check_tuples(list_of_tuples):
        """Method that checks if the passed tuple contains an optimizer in the
        first element, and a list of NmTensors in the second.
        """
        for tup in list_of_tuples:
            if not (isinstance(tup[0], torch.optim.Optimizer) and PtActions._check_all_tensors(tup[1])):
                return False
        return True

    @staticmethod
    def __module_export(module, output, d_format: DeploymentFormat, input_example=None, output_example=None):
        # Check if output already exists
        destination = Path(output)
        if destination.exists():
            raise FileExistsError(f"Destination {output} already exists. " f"Aborting export.")

        input_names = list(module.input_ports.keys())
        output_names = list(module.output_ports.keys())
        dynamic_axes = defaultdict(list)

        def __extract_dynamic_axes(port_name: str, ntype: NeuralType, dynamic_axes: defaultdict):
            if ntype.axes:
                for ind, axis in enumerate(ntype.axes):
                    if axis.kind == AxisKind.Batch or axis.kind == AxisKind.Time:
                        dynamic_axes[port_name].append(ind)

        # extract dynamic axes and remove unnecessary inputs/outputs
        # for input_ports
        for port_name, ntype in module.input_ports.items():
            if port_name in module._disabled_deployment_input_ports:
                input_names.remove(port_name)
                continue
            __extract_dynamic_axes(port_name, ntype, dynamic_axes)
        # for output_ports
        for port_name, ntype in module.output_ports.items():
            if port_name in module._disabled_deployment_output_ports:
                output_names.remove(port_name)
                continue
            __extract_dynamic_axes(port_name, ntype, dynamic_axes)

        if len(dynamic_axes) == 0:
            dynamic_axes = None

        # Make a deep copy of init parameters.
        init_params_copy = copy.deepcopy(module._init_params)

        # Reset standard instance field - making the file (probably) lighter.
        module._init_params = None
        module._placement = None
        module._factory = None
        module._device = None

        module.eval()
        try:
            # Remove NeMo-related things from the module
            # We need to change __call__ method. Note that this will change the
            # whole class, not just this object! Which is why we need to repair it
            # in the finally block
            __orig_call__ = type(module).__call__
            type(module).__call__ = torch.nn.Module.__call__

            if d_format == DeploymentFormat.TORCHSCRIPT:
                if input_example is None:
                    # Route 1 - via torch.jit.script
                    traced_m = torch.jit.script(module)
                    traced_m.save(output)
                else:
                    # Route 2 - via tracing
                    traced_m = torch.jit.trace(module, input_example)
                    traced_m.save(output)
            elif d_format == DeploymentFormat.ONNX or d_format == DeploymentFormat.TRTONNX:
                if input_example is None:
                    raise ValueError(f'Example input is None, but ONNX tracing was' f' attempted')
                if output_example is None:
                    if isinstance(input_example, tuple):
                        output_example = module.forward(*input_example)
                    else:
                        output_example = module.forward(input_example)
                with torch.jit.optimized_execution(True):
                    jitted_model = torch.jit.trace(module, input_example)

                torch.onnx.export(
                    jitted_model,
                    input_example,
                    output,
                    input_names=input_names,
                    output_names=output_names,
                    verbose=False,
                    export_params=True,
                    do_constant_folding=True,
                    dynamic_axes=dynamic_axes,
                    opset_version=11,
                    example_outputs=output_example,
                )
                # fn = output + ".readable"
                # with open(fn, 'w') as f:
                #     tempModel = onnx.load(output)
                #     onnx.save(tempModel, output + ".copy")
                #     onnx.checker.check_model(tempModel)
                #     pgraph = onnx.helper.printable_graph(tempModel.graph)
                #     f.write(pgraph)

            elif d_format == DeploymentFormat.PYTORCH:
                torch.save(module.state_dict(), output)
                with open(output + ".json", 'w') as outfile:
                    json.dump(init_params_copy, outfile)

            else:
                raise NotImplementedError(f"Not supported deployment format: {d_format}")
        except Exception as e:  # nopep8
            logging.error(f'module export failed for {module} ' f'with exception {e}')
        finally:
            type(module).__call__ = __orig_call__

    @staticmethod
    def deployment_export(module, output: str, d_format: DeploymentFormat, input_example=None, output_example=None):
        """Exports Neural Module instance for deployment.

        Args:
            module: neural module to export
            output (str): where export results should be saved
            d_format (DeploymentFormat): which deployment format to use
            input_example: sometimes tracing will require input examples
            output_example: Should match inference on input_example
            amp_max_loss_scale (float): Max value for amp loss scaling.
                Defaults to 2.0**24.
        """

        with torch.no_grad():
            PtActions.__module_export(
                module=module,
                output=output,
                d_format=d_format,
                input_example=input_example,
                output_example=output_example,
            )

    def train(
        self,
        tensors_to_optimize=None,
        training_graph=None,
        optimizer=None,
        optimization_params=None,
        callbacks: Optional[List[ActionCallback]] = None,
        lr_policy=None,
        batches_per_step=None,
        stop_on_nan_loss=False,
        synced_batchnorm=False,
        synced_batchnorm_groupsize=0,
        gradient_predivide=False,
        amp_max_loss_scale=2.0 ** 24,
    ):
        def _perform_on_step_start(callbacks, state):
            # TODO: Most of these checks can be relaxed since we enforce callbacks
            # to be a list of ActionCallback objects
            if callbacks is not None and isinstance(callbacks, List) and len(callbacks) > 0:
                for callback in callbacks:
                    if isinstance(callback, ActionCallback):
                        callback.on_iteration_start()
                    elif isinstance(callback, NeMoCallback):
                        callback.on_step_start(state)
                    else:
                        raise ValueError(
                            "Callback was not a child of ActionCallback nor NeMoCallback and was not understood"
                        )

        def _perform_on_step_end(callbacks, state):
            if callbacks is not None and isinstance(callbacks, List) and len(callbacks) > 0:
                for callback in callbacks:
                    if isinstance(callback, ActionCallback):
                        callback.on_iteration_end()
                    elif isinstance(callback, NeMoCallback):
                        callback.on_step_end(state)
                    else:
                        raise ValueError(
                            "Callback was not a child of ActionCallback nor NeMoCallback and was not understood"
                        )

        def _perform_on_action_start(callbacks, state):
            if callbacks is not None and isinstance(callbacks, List) and len(callbacks) > 0:
                for callback in callbacks:
                    if isinstance(callback, ActionCallback):
                        callback.on_action_start()
                    elif isinstance(callback, NeMoCallback):
                        callback.on_action_start(state)
                    else:
                        raise ValueError(
                            "Callback was not a child of ActionCallback nor NeMoCallback and was not understood"
                        )

        def _perform_on_action_end(callbacks, state):
            if callbacks is not None and isinstance(callbacks, List) and len(callbacks) > 0:
                for callback in callbacks:
                    if isinstance(callback, ActionCallback):
                        callback.on_action_end()
                    elif isinstance(callback, NeMoCallback):
                        callback.on_action_end(state)
                    else:
                        raise ValueError(
                            "Callback was not a child of ActionCallback nor NeMoCallback and was not understood"
                        )

        def _perform_on_epoch_start(callbacks, state):
            if callbacks is not None and isinstance(callbacks, List) and len(callbacks) > 0:
                for callback in callbacks:
                    if isinstance(callback, ActionCallback):
                        callback.on_epoch_start()
                    elif isinstance(callback, NeMoCallback):
                        callback.on_epoch_start(state)
                    else:
                        raise ValueError(
                            "Callback was not a child of ActionCallback nor NeMoCallback and was not understood"
                        )

        def _perform_on_epoch_end(callbacks, state):
            if callbacks is not None and isinstance(callbacks, List) and len(callbacks) > 0:
                for callback in callbacks:
                    if isinstance(callback, ActionCallback):
                        callback.on_epoch_end()
                    elif isinstance(callback, NeMoCallback):
                        callback.on_epoch_end(state)
                    else:
                        raise ValueError(
                            "Callback was not a child of ActionCallback nor NeMoCallback and was not understood"
                        )

        def _perform_on_batch_start(callbacks, state):
            if callbacks is not None and isinstance(callbacks, List) and len(callbacks) > 0:
                for callback in callbacks:
                    if isinstance(callback, ActionCallback):
                        continue
                    elif isinstance(callback, NeMoCallback):
                        callback.on_batch_start(state)
                    else:
                        raise ValueError(
                            "Callback was not a child of ActionCallback nor NeMoCallback and was not understood"
                        )

        def _perform_on_batch_end(callbacks, state):
            if callbacks is not None and isinstance(callbacks, List) and len(callbacks) > 0:
                for callback in callbacks:
                    if isinstance(callback, ActionCallback):
                        continue
                    elif isinstance(callback, NeMoCallback):
                        callback.on_batch_end(state)
                    else:
                        raise ValueError(
                            "Callback was not a child of ActionCallback nor NeMoCallback and was not understood"
                        )

        def _init_callbacks(callbacks, action):
            if callbacks is not None and isinstance(callbacks, List) and len(callbacks) > 0:
                for callback in callbacks:
                    if isinstance(callback, ActionCallback):
                        callback.action = action

        def _update_callbacks(callbacks=None, registered_tensors=None, final_loss=None):
            # if self.local_rank is None or self.local_rank == 0:
            if callbacks is not None and isinstance(callbacks, List) and len(callbacks) > 0:
                for callback in callbacks:
                    if isinstance(callback, ActionCallback):
                        callback._registered_tensors = registered_tensors
                    else:  # For now, we can use the old callback function. In the future we should improve this
                        registered_tensors["loss"] = final_loss

        def get_state(action: 'PtAction'):
            """Helper function used to create a state for callbacks
            """

            class StateWrapper(dict):
                def __init__(self, action):
                    """A class that wraps a dictionary but adds the functions: restore_state_from and save_state_to
                    which are helper functions for CheckpointCallback to use.
                    The StateWrapper is a dictionary that contains the following mapping:
                        "step" (int): the current training step
                        "epoch" (int): the current epoch step
                        "local_rank" (int): the local rank that the process is running on
                        "global_rank" (int): the global rank that the process is running on
                        "optimizers" (list): a list of optimizers defined during the training process
                        "tensors" (TrainingState): A TrainingState object that can be used to access tensor values
                    """
                    self.action = action
                    super().__init__(
                        {
                            "step": action.step,
                            "tensors": action._training_state,
                            "epoch": action.epoch,
                            "local_rank": action.local_rank,
                            "global_rank": action.global_rank,
                            "optimizers": action.optimizers,
                        }
                    )

                def restore_state_from(self, path):
                    if os.path.isfile(path):
                        # map_location could be cuda:<device_id> but cpu seems to be more
                        # general since we are also saving step and epoch
                        # load_state_dict should move the variables to the relevant device
                        checkpoint = torch.load(path, map_location="cpu")
                        action.step = checkpoint["step"]
                        self["step"] = action.step
                        epoch = checkpoint.get("epoch", None)
                        if epoch is None:
                            epoch = checkpoint.get("epoch_num", None)
                        if epoch is None:
                            raise ValueError("Epoch was not found in the trainer checkpoint")
                        action.epoch = epoch
                        self["epoch"] = action.epoch
                        if checkpoint["optimizer_state"]:
                            for opt, opt_chkpt in zip(self["optimizers"], checkpoint["optimizer_state"]):
                                opt.load_state_dict(opt_chkpt)
                    else:
                        raise FileNotFoundError("Could not find checkpoint file: {0}".format(path))

                def save_state_to(self, path):
                    state = {
                        "step": self["step"],
                        "epoch": self["epoch"],
                        "optimizer_state": [opt.state_dict() for opt in self["optimizers"]],
                    }
                    torch.save(state, path)

            return StateWrapper(action)

        if self._train_called:
            logging.warning(
                "You called train twice. Please note that we do not support calling training twice in one script if "
                "amp or ddp is used. If you wish to call train twice, you need to run "
                "`nemo.utils.app_state.AppState().modules.clear(); neural_factory.reset_trainer()` and then "
                "reinstantiate all Neural Modules prior to calling train()"
            )
        self._train_called = True

        self._training_state = TrainingState(self)
        # Analyse the arguments passed to train.
        if tensors_to_optimize is not None and training_graph is not None:
            raise ValueError("Cannot pass both `tensors_to_optimize` and `training_graph` to the train() function")
        # if tensors_to_optimize is None and training_graph is None:
        #    raise ValueError(
        #        "One of the `tensors_to_optimize` or `training_graph` values must be passed to the train() function"
        #    )
        # Finally, unify.
        if training_graph is not None:
            # Get device - from NF.
            device_type = NeuralModuleFactory.get_default_factory().placement
            # Move the graph to device.
            training_graph.to(device_type=device_type)
            # To keep the "compatibility with old NeMo": get output tensors.
            tensors_to_optimize = training_graph.outputs.tensor_list

        if gradient_predivide:
            logging.error(
                "gradient_predivide is currently disabled, and is under consideration for removal in future versions. "
                "If this functionality is needed, please raise a github issue."
            )
        if not optimization_params:
            optimization_params = {}
        num_epochs = optimization_params.get("num_epochs", None)
        max_steps = optimization_params.get("max_steps", None)
        if num_epochs is None and max_steps is None:
            raise ValueError("You must specify either max_steps or num_epochs")
        grad_norm_clip = optimization_params.get('grad_norm_clip', None)

        if batches_per_step is None:
            batches_per_step = 1
        # this is necessary because we average gradients over batch
        bps_scale = torch.FloatTensor([1.0 / batches_per_step]).squeeze()

        if tensors_to_optimize is None:
            # This is Evaluation Mode
            _init_callbacks(callbacks, self)
            # Do action start callbacks
            _perform_on_action_end(callbacks, get_state(self))
            return
        # Check if tensors_to_optimize is just a list of NmTensors
        elif tensors_to_optimize is not None and (
            isinstance(tensors_to_optimize[0], NmTensor) and PtActions._check_all_tensors(tensors_to_optimize)
        ):
            # Parse graph into a topologically sorted sequence of neural
            # modules' calls
            (opt_call_chain, t_dataset,) = self.__get_top_sorted_modules_and_dataloader(hook=tensors_to_optimize)
            # Extract trainable weights which will be optimized
            params_list = [
                p[0].parameters() for p in opt_call_chain if isinstance(p[0], TrainableNM) or p[0].is_trainable()
            ]
            params_to_optimize = itertools.chain(*params_list)

            # Setup optimizer instance. By default it is SGD
            optimizer_instance = None
            optimizer_class = None
            if isinstance(optimizer, str):
                optimizer_class = optimizer
            elif isinstance(optimizer, torch.optim.Optimizer):
                optimizer_instance = optimizer
            else:
                raise ValueError("optimizer was not understood")
            optimizer = self.__setup_optimizer(
                optimizer_instance=optimizer_instance,
                optimizer_class=optimizer_class,
                optimization_params=optimization_params,
                params_to_optimize=params_to_optimize,
            )

            training_loop = [(optimizer, tensors_to_optimize, opt_call_chain)]

            self._optimizers.append(optimizer)
            assert len(self._optimizers) == 1, (
                "There was more than one optimizer, was create_optimizer() called before train()? Are you calling "
                "train() twice in one script, If so you need to call NeuralModuleFactory.reset_trainer() first."
            )

        elif PtActions._check_tuples(tensors_to_optimize):
            if batches_per_step != 1:
                raise ValueError("Gradient accumlation with multiple optimizers is not supported")
            datasets = []
            training_loop = []
            for step in tensors_to_optimize:
                (step_call_chain, dataset,) = self.__get_top_sorted_modules_and_dataloader(hook=step[1])
                datasets.append(dataset)
                training_loop.append((step[0], step[1], step_call_chain))

            t_dataset = datasets[0]
            for dataset in datasets:
                if type(dataset) is not type(t_dataset):
                    raise ValueError("There were two training datasets, we only support 1.")
        else:
            raise ValueError("tensors_to_optimize was not understood")

        logging_callchain = None
        # callbacks setup
        if callbacks is not None:
            for callback in callbacks:
                if not isinstance(callback, ActionCallback) and not isinstance(callback, NeMoCallback):
                    raise ValueError(
                        "A callback was received that was not a child of ActionCallback nor a child of NeMoCallback"
                    )
                elif isinstance(callback, SimpleLossLoggerCallback):
                    if logging_callchain:
                        raise ValueError("We only support one logger callback but more than one were found")
                    logger_step_freq = callback._step_freq
                    logging_tensors = callback.tensors
                    all_tensors = logging_tensors
                    for step in training_loop:
                        all_tensors = all_tensors + step[1]
                    (logging_callchain, _,) = self.__get_top_sorted_modules_and_dataloader(hook=all_tensors)

        # Intialize Amp if needed
        if self._optim_level in AmpOptimizations:
            # Store mapping of self.optimizers to optimizer in callchain
            training_loop_opts = []
            for opt in training_loop:
                training_loop_opts.append(self._optimizers.index(opt[0]))
            self._optimizers = self.__initialize_amp(
                optimizer=self._optimizers,
                optim_level=self._optim_level,
                amp_max_loss_scale=amp_max_loss_scale,
                amp_min_loss_scale=optimization_params.get('amp_min_loss_scale', 1.0),
            )
            # Use stored mapping to map amp_init opts to training loop
            for i, step in enumerate(training_loop):
                training_loop[i] = (
                    self._optimizers[training_loop_opts[i]],
                    step[1],
                    step[2],
                )

        dataNM = training_loop[0][2][0][0]
        placement_gpu = dataNM.placement == DeviceType.AllGpu
        if placement_gpu:
            logging.info("Doing distributed training")
            if t_dataset is not None:
                train_sampler = None
                if not isinstance(t_dataset, torch.utils.data.IterableDataset):
                    train_sampler = torch.utils.data.distributed.DistributedSampler(
                        dataset=t_dataset, shuffle=dataNM.shuffle
                    )
                dataloader_params = {
                    'dataset': t_dataset,
                    'sampler': train_sampler,
                    'num_workers': dataNM.num_workers,
                    'batch_size': dataNM.batch_size,
                    'shuffle': False,
                    'pin_memory': dataNM.pin_memory,
                }
                if hasattr(dataNM, 'collate_fn'):
                    dataloader_params['collate_fn'] = dataNM.collate_fn
                train_dataloader = torch.utils.data.DataLoader(**dataloader_params)
            else:
                train_dataloader = dataNM.data_iterator
                if hasattr(train_dataloader, 'sampler'):
                    train_sampler = train_dataloader.sampler
                else:
                    train_sampler = None

            self.ddp_initialized = True
            module_list = [mod.name for mod in AppState().modules]
            module_list = sorted(module_list)
            for module_name in module_list:
                module = AppState().modules[module_name]
                key = module.unique_instance_id
                num_trainable_weights = module.num_weights
                self.ddp_module_dict[key] = module
                if not isinstance(module, DDP) and isinstance(module, torch.nn.Module) and num_trainable_weights > 0:
                    # Per pytorch docs, convert sync bn prior to DDP
                    if synced_batchnorm:
                        world_size = dist.get_world_size()
                        sync_batchnorm_group = None
                        if synced_batchnorm_groupsize > 0:
                            if world_size % synced_batchnorm_groupsize != 0:
                                raise ValueError(
                                    f"Synchronized batch norm group size ({synced_batchnorm_groupsize}) must be 0"
                                    f" or divide total number of GPUs ({world_size})."
                                )
                            # Find ranks of other nodes in the same batchnorm group
                            rank = torch.distributed.get_rank()
                            group = rank // synced_batchnorm_groupsize
                            group_rank_ids = range(
                                group * synced_batchnorm_groupsize, (group + 1) * synced_batchnorm_groupsize
                            )
                            sync_batchnorm_group = torch.distributed.new_group(group_rank_ids)

                        module = nn.SyncBatchNorm.convert_sync_batchnorm(module, process_group=sync_batchnorm_group)

                    # By default, disable broadcast_buffers. This disables batch norm synchronization on forward
                    # pass
                    module = DDP(
                        module, device_ids=[self.local_rank], broadcast_buffers=False, find_unused_parameters=True
                    )
                    self.ddp_module_dict[key] = module

        # single GPU/CPU training
        else:
            if t_dataset is not None:
                train_sampler = None
                dataloader_params = {
                    'dataset': t_dataset,
                    'sampler': None,
                    'num_workers': dataNM.num_workers,
                    'batch_size': dataNM.batch_size,
                    'shuffle': dataNM.shuffle,
                    'pin_memory': dataNM.pin_memory,
                }
                if hasattr(dataNM, 'collate_fn'):
                    dataloader_params['collate_fn'] = dataNM.collate_fn

                train_dataloader = torch.utils.data.DataLoader(**dataloader_params)
            else:
                train_dataloader = dataNM.data_iterator
                train_sampler = None

        _init_callbacks(callbacks, self)
        # Do action start callbacks
        _perform_on_action_start(callbacks, get_state(self))

        # MAIN TRAINING LOOP
        # iteration over epochs
        while num_epochs is None or self.epoch < num_epochs:
            if train_sampler is not None:
                train_sampler.set_epoch(self.epoch)
            if max_steps is not None and self.step >= max_steps:
                break

            # Register epochs start with callbacks
            _perform_on_epoch_start(callbacks, get_state(self))

            # iteration over batches in epoch
            batch_counter = 0
            for _, data in enumerate(train_dataloader, 0):
                if max_steps is not None and self.step >= max_steps:
                    break

                if batch_counter == 0:
                    # Started step, zero gradients
                    curr_optimizer = training_loop[self.step % len(training_loop)][0]
                    curr_optimizer.zero_grad()
                    # Register iteration start with callbacks
                    _perform_on_step_start(callbacks, get_state(self))

                # Perform batch start callbacks
                _perform_on_batch_start(callbacks, get_state(self))

                # set learning rate policy
                if lr_policy is not None:
                    adjusted_lr = lr_policy(optimization_params["lr"], self.step, self.epoch)
                    for param_group in curr_optimizer.param_groups:
                        param_group["lr"] = adjusted_lr

                # TODO: Remove below loop when ActionCallback is removed
                if callbacks is not None:
                    for callback in callbacks:
                        if isinstance(callback, ActionCallback):
                            callback.learning_rate = curr_optimizer.param_groups[0]['lr']

                # registered_tensors will contain created tensors
                # named by output port and uuid of module which created them
                # Get and properly name tensors returned by data layer
                curr_call_chain = training_loop[self.step % len(training_loop)][2]
                dl_device = curr_call_chain[0][0]._device
                if logging_callchain and self.step % logger_step_freq == 0:
                    curr_call_chain = logging_callchain
                tensors = []
                if isinstance(data, torch.Tensor):
                    data = (data,)
                for d in data:
                    if isinstance(d, torch.Tensor):
                        tensors.append(d.to(dl_device))
                    else:
                        tensors.append(d)

                for t, d in zip(curr_call_chain[0][2].values(), tensors):
                    if t is not None:
                        self._training_state.set_tensor(t, d)
                disable_allreduce = batch_counter < (batches_per_step - 1)
                self.__nm_graph_forward_pass(
                    call_chain=curr_call_chain, registered_tensors=self._training_state.tensor_dict,
                )

                curr_tensors_to_optimize = training_loop[self.step % len(training_loop)][1]
                final_loss = 0
                for tensor in curr_tensors_to_optimize:
                    final_loss += self._training_state.tensor_dict[tensor.unique_name]

                # Check for NaN/inf loss (across workers if applicable)
                loss_nan_inf_checker = final_loss.clone()
                if placement_gpu:
                    dist.all_reduce(loss_nan_inf_checker, torch.distributed.ReduceOp.MAX)
                if torch.isnan(loss_nan_inf_checker).any() or torch.isinf(loss_nan_inf_checker).any():
                    if stop_on_nan_loss:
                        raise ValueError('Loss is NaN or inf - exiting')
                    if self._optim_level in AmpOptimizations and self._optim_level != Optimization.mxprO0:
                        logging.warning('Loss is NaN or inf.')
                    else:
                        # Skip this step across workers if loss is NaN/inf and using fp32
                        logging.warning('Loss is NaN or inf. Skipping update.')
                        self._training_state.clear_dict()  # Clear state dict here
                        continue

                if self._optim_level in AmpOptimizations and self._optim_level != Optimization.mxprO0:
                    with amp.scale_loss(final_loss, curr_optimizer, delay_unscale=disable_allreduce) as scaled_loss:
                        if disable_allreduce:
                            with ExitStack() as stack:
                                for mod in self.get_DDP_modules(curr_call_chain):
                                    stack.enter_context(mod.no_sync())
                                scaled_loss.backward(bps_scale.to(scaled_loss.get_device()))
                        else:
                            scaled_loss.backward(bps_scale.to(scaled_loss.get_device()))
                # no AMP optimizations needed
                else:
                    # multi-GPU, float32
                    if self._local_rank is not None:
                        if disable_allreduce:
                            with ExitStack() as stack:
                                for mod in self.get_DDP_modules(curr_call_chain):
                                    stack.enter_context(mod.no_sync())
                                final_loss.backward(bps_scale.to(final_loss.get_device()))
                        else:
                            final_loss.backward(bps_scale.to(final_loss.get_device()))
                    # single device (CPU or GPU)
                    else:
                        # Fix (workaround?) enabling to backpropagate gradients on CPUs.
                        if final_loss.get_device() < 0:
                            final_loss.backward(bps_scale)
                        else:
                            final_loss.backward(bps_scale.to(final_loss.get_device()))

                # Register batch end with callbacks
                _update_callbacks(
                    callbacks, registered_tensors=self._training_state.tensor_dict, final_loss=final_loss
                )
                # Perform batch end callbacks
                _perform_on_batch_end(callbacks, get_state(self))

                batch_counter += 1
                if batch_counter == batches_per_step:
                    # Ended step. Do optimizer update
                    if grad_norm_clip is not None:
                        torch.nn.utils.clip_grad_norm_(master_params(curr_optimizer), grad_norm_clip)
                    curr_optimizer.step()
                    batch_counter = 0
                    _perform_on_step_end(callbacks, get_state(self))
                    self.step += 1
                self._training_state.clear_dict()
            # End of epoch for loop
            # Register epochs end with callbacks
            _perform_on_epoch_end(callbacks, get_state(self))
            self.epoch += 1
        _perform_on_action_end(callbacks, get_state(self))

    def infer(
        self,
        tensors,
        checkpoint_dir=None,
        ckpt_pattern='',
        verbose=True,
        cache=False,
        use_cache=False,
        offload_to_cpu=True,
        modules_to_restore=None,
    ):
        """See NeuralModuleFactory.infer()
        """

        call_chain, _ = self.__get_top_sorted_modules_and_dataloader(hook=tensors)
        if checkpoint_dir:
            # Find all modules that need to be restored
            if modules_to_restore is None:
                modules_to_restore = []
                modules_to_restore_name = []
                for op in call_chain:
                    if op[0].num_weights > 0:
                        modules_to_restore.append(op[0])

            if not isinstance(modules_to_restore, list):
                modules_to_restore = [modules_to_restore]
            modules_to_restore_name = []
            for mod in modules_to_restore:
                if not isinstance(mod, NeuralModule):
                    raise ValueError("Found something that was not a Neural Module inside modules_to_restore")
                elif mod.num_weights == 0:
                    raise ValueError("Found a Neural Module with 0 weights inside modules_to_restore")
                modules_to_restore_name.append(str(mod))

            module_checkpoints = get_checkpoint_from_dir(modules_to_restore_name, checkpoint_dir, ckpt_pattern)

            for mod, checkpoint in zip(modules_to_restore, module_checkpoints):
                logging.info(f"Restoring {mod} from {checkpoint}")
                mod.restore_from(checkpoint, self._local_rank)

        # Init Amp
        if (
            self._optim_level in AmpOptimizations
            and self._optim_level != Optimization.mxprO0
            and not self.amp_initialized
        ):
            pt_modules = []
            for i in range(len(call_chain)):
                if isinstance(call_chain[i][0], nn.Module):
                    pt_modules.append(call_chain[i][0])
                elif isinstance(call_chain[i][0], TrainableNeuralModuleWrapper):
                    pt_modules.append(call_chain[i][0]._pt_module)

            amp.initialize(
                min_loss_scale=1.0, models=pt_modules, optimizers=None, opt_level=AmpOptimizations[self._optim_level],
            )
            self.amp_initialized = True

        # Run infer
        return self._infer(
            tensors_to_return=tensors,
            verbose=verbose,
            cache=cache,
            use_cache=use_cache,
            offload_to_cpu=offload_to_cpu,
        )

    def get_DDP_modules(self, call_chain):
        modules = []
        for ind in range(1, len(call_chain)):
            m_id = call_chain[ind][0].unique_instance_id
            module = self.ddp_module_dict[m_id]
            if isinstance(module, DDP):
                modules.append(module)

        return modules



# -*- coding: utf-8 -*-
# =============================================================================
# Copyright (c) 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from weakref import WeakSet


class ObjectRegistry(WeakSet):
    """
        Registry used for storing references to objects, generating unique names and monitoring their `uniqueness`.
    """

    def __init__(self, base_type_name):
        """
            Stores base type name.
        """
        super().__init__()
        self._base_type_name = base_type_name

    def register(self, new_obj, name: str) -> str:
        """
            Registers a new object using the provided name.
            If name is none - generates new unique name.

            Args:
                new_obj: An object to be registered.
                name: A "proposition" for the object name.

            Returns:
                A unique name (proposition or newly generated name).
        """

        # Check if object is already in a set.
        if new_obj in self:
            # Return its name.
            return new_obj.name

        # Check object name.
        if name is None:
            # Generate a new, unique name.
            unique_name = self.__generate_unique_name(new_obj)
        else:
            # Check if name is unique.
            if self.has(name):
                raise NameError("A {} with name `{}` already exists!".format(name, name))
            # Ok, it is unique.
            unique_name = name

        # Finally, add object to the set.
        self.add(new_obj)

        # Return the name.
        return unique_name

    def has(self, name: str) -> bool:
        """
            Check if registry stores object with a given name.

            Args:
                name: name of the object to be found in the registry.
        """
        for obj in self:
            if obj.name == name:
                return True
        # Else:
        return False

    def __generate_unique_name(self, new_obj) -> str:
        """
            Generates a new unique name by adding postfix (number) to base name.

            Args:
                new_obj: An object to be registered.

            Returns:
                A generated unique name.
        """
        # Iterate through numbers.
        postfix = 0
        # Get type name.
        base_type_name = (type(new_obj).__name__).lower()
        while True:
            # Generate name.
            new_name = base_type_name + str(postfix)
            # Check uniqueneess.
            if not self.has(new_name):
                # Ok, got a unique name!
                break
            # Increment index.
            postfix += 1
        return new_name

    def __getitem__(self, key: str):
        """
        Object getter function.

        Args:
            key: Object name.

        Returns:
            Object associated with the key.
        """
        # Search for an object with a given name.
        for obj in self:
            # Retrieve object
            if obj.name == key:
                return obj
        # Else: seems that there is no object with that name.
        raise KeyError("A {} with name `{}` don't exists!".format(self._base_type_name, key))

    def __eq__(self, other):
        """
            Checks if two registers have the same content.

            Args:
                other: The second registry object.
        """
        if not isinstance(other, WeakSet):
            return False
        return super().__eq__(other)

    def summary(self) -> str:
        """
            Returns:
                A summary of the objects on the list.
        """
        # Line "decorator".
        summary = "\n" + 113 * '=' + "\n"
        summary += "Registry of {}s:\n".format(self._base_type_name)
        for obj in self:
            summary += " * {} ({})\n".format(obj.name, type(obj).__name__)
        # Line "decorator".
        summary += 113 * '='
        return summary



# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from nemo.backends.pytorch.common.losses import CrossEntropyLossNM
from nemo.collections.asr import models
from nemo.collections.asr.audio_preprocessing import *
from nemo.collections.asr.beam_search_decoder import BeamSearchDecoderWithLM
from nemo.collections.asr.contextnet import ContextNetDecoderForCTC, ContextNetEncoder
from nemo.collections.asr.data_layer import (
    AudioToSpeechLabelDataLayer,
    AudioToTextDataLayer,
    KaldiFeatureDataLayer,
    TarredAudioToTextDataLayer,
    TranscriptDataLayer,
)
from nemo.collections.asr.greedy_ctc_decoder import GreedyCTCDecoder
from nemo.collections.asr.jasper import (
    JasperDecoderForClassification,
    JasperDecoderForCTC,
    JasperDecoderForSpkrClass,
    JasperEncoder,
)
from nemo.collections.asr.las.misc import JasperRNNConnector
from nemo.collections.asr.losses import CTCLossNM

__all__ = [
    'AudioToTextDataLayer',
    'TarredAudioToTextDataLayer',
    'AudioToSpeechLabelDataLayer',
    'AudioPreprocessing',
    'AudioPreprocessor',
    'AudioToMFCCPreprocessor',
    'AudioToMelSpectrogramPreprocessor',
    'AudioToSpectrogramPreprocessor',
    'CropOrPadSpectrogramAugmentation',
    'MultiplyBatch',
    'SpectrogramAugmentation',
    'KaldiFeatureDataLayer',
    'TranscriptDataLayer',
    'GreedyCTCDecoder',
    'BeamSearchDecoderWithLM',
    'JasperEncoder',
    'JasperDecoderForCTC',
    'JasperDecoderForClassification',
    'JasperDecoderForSpkrClass',
    'JasperRNNConnector',
    'ContextNetEncoder',
    'ContextNetDecoderForCTC',
    'CTCLossNM',
    'CrossEntropyLossNM',
]



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import json

from nemo import logging
from nemo.collections.nlp.nm.trainables.common.huggingface.huggingface_utils import *

try:
    __megatron_utils_satisfied = True
    from nemo.collections.nlp.nm.trainables.common.megatron.megatron_bert_nm import MegatronBERT
    from nemo.collections.nlp.nm.trainables.common.megatron.megatron_utils import *

except Exception as e:
    logging.error('Failed to import Megatron Neural Module and utils: `{}` ({})'.format(str(e), type(e)))
    __megatron_utils_satisfied = False


__all__ = ['get_pretrained_lm_models_list', 'get_pretrained_lm_model']


def get_pretrained_lm_models_list():
    '''
    Returns the list of support pretrained models
    '''
    if __megatron_utils_satisfied:
        return get_megatron_lm_models_list() + get_huggingface_lm_models_list()
    else:
        return get_huggingface_lm_models_list()


def get_pretrained_lm_model(pretrained_model_name, config=None, vocab=None, checkpoint=None):
    '''
    Returns pretrained model
    Args:
        pretrained_model_name (str): pretrained model name, for example, bert-base-uncased.
            See the full list by calling get_pretrained_lm_models_list()
        config (str): path to the model configuration file
        vocab (str): path to the vocabulary file used during model training 
        checkpoint (str): path to the pretrained model checkpoint
    Returns:
        Pretrained model (NM)
    '''
    if pretrained_model_name in get_huggingface_lm_models_list():
        model = get_huggingface_lm_model(bert_config=config, pretrained_model_name=pretrained_model_name)
    elif __megatron_utils_satisfied and pretrained_model_name in get_megatron_lm_models_list():
        if pretrained_model_name == 'megatron-bert-cased' or pretrained_model_name == 'megatron-bert-uncased':
            if not (config and checkpoint):
                raise ValueError(f'Config file and pretrained checkpoint required for {pretrained_model_name}')
        if not config:
            config = get_megatron_config_file(pretrained_model_name)
        if isinstance(config, str):
            with open(config) as f:
                config = json.load(f)
        if not vocab:
            vocab = get_megatron_vocab_file(pretrained_model_name)
        if not checkpoint:
            checkpoint = get_megatron_checkpoint(pretrained_model_name)
        model = MegatronBERT(
            model_name=pretrained_model_name,
            vocab_file=vocab,
            hidden_size=config['hidden-size'],
            num_attention_heads=config['num-attention-heads'],
            num_layers=config['num-layers'],
            max_seq_length=config['max-seq-length'],
        )
    else:
        raise ValueError(f'{pretrained_model_name} is not supported')

    if checkpoint:
        model.restore_from(checkpoint)
        logging.info(f"{pretrained_model_name} model restored from {checkpoint}")
    return model



# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from nemo.collections.tts.data_layers import AudioDataLayer
from nemo.collections.tts.fastspeech_modules import *
from nemo.collections.tts.fastspeech_modules import __all__ as fastspeech__all__
from nemo.collections.tts.parts.helpers import *
from nemo.collections.tts.parts.helpers import __all__ as helpers__all__
from nemo.collections.tts.tacotron2_modules import *
from nemo.collections.tts.tacotron2_modules import __all__ as tacotron2__all__
from nemo.collections.tts.talknet_modules import *
from nemo.collections.tts.talknet_modules import __all__ as fasterspeech__all__
from nemo.collections.tts.waveglow_modules import *
from nemo.collections.tts.waveglow_modules import __all__ as waveglow__all__

__all__ = (
    ["AudioDataLayer"] + helpers__all__ + tacotron2__all__ + waveglow__all__ + fastspeech__all__ + fasterspeech__all__
)



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2018 The Google AI Language Team Authors and
# The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import collections
import json
import os
import pickle

import numpy as np
import torch
from torch.utils.data import Dataset
from tqdm import tqdm

from nemo import logging
from nemo.collections.nlp.data.datasets.datasets_utils.data_preprocessing import is_whitespace
from nemo.collections.nlp.data.datasets.datasets_utils.datasets_processing import DataProcessor
from nemo.collections.nlp.data.datasets.qa_squad_dataset.qa_squad_processing import convert_examples_to_features
from nemo.collections.nlp.metrics.squad_metrics import (
    _get_best_indexes,
    apply_no_ans_threshold,
    exact_match_score,
    f1_score,
    find_all_best_thresh,
    get_final_text,
    make_eval_dict,
    merge_eval,
)
from nemo.collections.nlp.utils.data_utils import normalize_answer
from nemo.collections.nlp.utils.functional_utils import _compute_softmax

__all__ = ['SquadDataset']

"""
Utility functions for Question Answering NLP tasks
Some parts of this code were adapted from the HuggingFace library at
https://github.com/huggingface/transformers
"""


class SquadDataset(Dataset):
    """
    Creates SQuAD dataset for Question Answering.

    Args:
        data_file (str): train.*.json eval.*.json or test.*.json.
        tokenizer (obj): Tokenizer object, e.g. NemoBertTokenizer.
        version_2_with_negative (bool): True if training should allow
            unanswerable questions.
        doc_stride (int): When splitting up a long document into chunks,
            how much stride to take between chunks.
        max_query_length (iny): All training files which have a duration less
            than min_duration are dropped. Can't be used if the `utt2dur` file
            does not exist. Defaults to None.
        max_seq_length (int): All training files which have a duration more
            than max_duration are dropped. Can't be used if the `utt2dur` file
            does not exist. Defaults to None.
        mode (str): Use "train", "eval" or "test" to define between
            training and evaluation.
    """

    def __init__(
        self,
        data_file,
        tokenizer,
        doc_stride,
        max_query_length,
        max_seq_length,
        version_2_with_negative,
        mode,
        use_cache,
    ):
        self.tokenizer = tokenizer
        self.version_2_with_negative = version_2_with_negative
        self.processor = SquadProcessor(data_file=data_file, mode=mode)
        self.mode = mode
        if mode not in ["eval", "train", "test"]:
            raise ValueError(f"mode should be either 'train', 'eval', or 'test' but got {mode}")
        self.examples = self.processor.get_examples()

        tokenizer_type = type(tokenizer.tokenizer).__name__
        vocab_size = getattr(tokenizer, "vocab_size", 0)
        cached_features_file = (
            data_file
            + '_cache'
            + '_{}_{}_{}_{}_{}_{}'.format(
                mode, tokenizer_type, str(vocab_size), str(max_seq_length), str(doc_stride), str(max_query_length)
            )
        )

        if use_cache and os.path.exists(cached_features_file):
            logging.info(f"loading from {cached_features_file}")
            with open(cached_features_file, "rb") as reader:
                self.features = pickle.load(reader)
        else:
            self.features = convert_examples_to_features(
                examples=self.examples,
                tokenizer=tokenizer,
                max_seq_length=max_seq_length,
                doc_stride=doc_stride,
                max_query_length=max_query_length,
                has_groundtruth=mode != "test",
            )

            if use_cache:
                master_device = not torch.distributed.is_initialized() or torch.distributed.get_rank() == 0
                if master_device:
                    logging.info("  Saving train features into cached file %s", cached_features_file)
                    with open(cached_features_file, "wb") as writer:
                        pickle.dump(self.features, writer)

    def __len__(self):
        return len(self.features)

    def __getitem__(self, idx):
        feature = self.features[idx]
        if self.mode == "test":
            return (
                np.array(feature.input_ids),
                np.array(feature.segment_ids),
                np.array(feature.input_mask),
                np.array(feature.unique_id),
            )
        else:
            return (
                np.array(feature.input_ids),
                np.array(feature.segment_ids),
                np.array(feature.input_mask),
                np.array(feature.unique_id),
                np.array(feature.start_position),
                np.array(feature.end_position),
            )

    def get_predictions(
        self,
        unique_ids,
        start_logits,
        end_logits,
        n_best_size,
        max_answer_length,
        do_lower_case,
        version_2_with_negative,
        null_score_diff_threshold,
    ):
        example_index_to_features = collections.defaultdict(list)

        unique_id_to_pos = {}
        for index, unique_id in enumerate(unique_ids):
            unique_id_to_pos[unique_id] = index

        for feature in self.features:
            example_index_to_features[feature.example_index].append(feature)

        _PrelimPrediction = collections.namedtuple(
            "PrelimPrediction", ["feature_index", "start_index", "end_index", "start_logit", "end_logit"]
        )

        all_predictions = collections.OrderedDict()
        all_nbest_json = collections.OrderedDict()
        scores_diff_json = collections.OrderedDict()

        for (example_index, example) in enumerate(self.examples):

            features = example_index_to_features[example_index]

            prelim_predictions = []
            # keep track of the minimum score of null start+end of position 0
            # large and positive
            score_null = 1000000
            # the paragraph slice with min null score
            min_null_feature_index = 0
            # start logit at the slice with min null score
            null_start_logit = 0
            # end logit at the slice with min null score
            null_end_logit = 0
            for (feature_index, feature) in enumerate(features):
                pos = unique_id_to_pos[feature.unique_id]
                start_indexes = _get_best_indexes(start_logits[pos], n_best_size)
                end_indexes = _get_best_indexes(end_logits[pos], n_best_size)
                # if we could have irrelevant answers,
                # get the min score of irrelevant
                if version_2_with_negative:
                    feature_null_score = start_logits[pos][0] + end_logits[pos][0]
                    if feature_null_score < score_null:
                        score_null = feature_null_score
                        min_null_feature_index = feature_index
                        null_start_logit = start_logits[pos][0]
                        null_end_logit = end_logits[pos][0]
                for start_index in start_indexes:
                    for end_index in end_indexes:
                        # We could hypothetically create invalid predictions,
                        # e.g., predict that the start of the span is in the
                        # question. We throw out all invalid predictions.
                        if start_index >= len(feature.tokens):
                            continue
                        if end_index >= len(feature.tokens):
                            continue
                        if start_index not in feature.token_to_orig_map:
                            continue
                        if end_index not in feature.token_to_orig_map:
                            continue
                        if not feature.token_is_max_context.get(start_index, False):
                            continue
                        if end_index < start_index:
                            continue
                        length = end_index - start_index + 1
                        if length > max_answer_length:
                            continue
                        prelim_predictions.append(
                            _PrelimPrediction(
                                feature_index=feature_index,
                                start_index=start_index,
                                end_index=end_index,
                                start_logit=start_logits[pos][start_index],
                                end_logit=end_logits[pos][end_index],
                            )
                        )

            if version_2_with_negative:
                prelim_predictions.append(
                    _PrelimPrediction(
                        feature_index=min_null_feature_index,
                        start_index=0,
                        end_index=0,
                        start_logit=null_start_logit,
                        end_logit=null_end_logit,
                    )
                )
            prelim_predictions = sorted(prelim_predictions, key=lambda x: (x.start_logit + x.end_logit), reverse=True)

            _NbestPrediction = collections.namedtuple("NbestPrediction", ["text", "start_logit", "end_logit"])

            seen_predictions = {}
            nbest = []
            for pred in prelim_predictions:
                if len(nbest) >= n_best_size:
                    break
                feature = features[pred.feature_index]
                if pred.start_index > 0:  # this is a non-null prediction
                    tok_tokens = feature.tokens[pred.start_index : (pred.end_index + 1)]
                    orig_doc_start = feature.token_to_orig_map[pred.start_index]
                    orig_doc_end = feature.token_to_orig_map[pred.end_index]
                    orig_tokens = example.doc_tokens[orig_doc_start : (orig_doc_end + 1)]
                    tok_text = " ".join(tok_tokens)

                    # De-tokenize WordPieces that have been split off.
                    tok_text = tok_text.replace(" ##", "")
                    tok_text = tok_text.replace("##", "")

                    # Clean whitespace
                    tok_text = tok_text.strip()
                    tok_text = " ".join(tok_text.split())
                    orig_text = " ".join(orig_tokens)

                    final_text = get_final_text(tok_text, orig_text, do_lower_case)
                    if final_text in seen_predictions:
                        continue

                    seen_predictions[final_text] = True
                else:
                    final_text = ""
                    seen_predictions[final_text] = True

                nbest.append(_NbestPrediction(text=final_text, start_logit=pred.start_logit, end_logit=pred.end_logit))
            # if we didn't include the empty option in the n-best, include it
            if version_2_with_negative:
                if "" not in seen_predictions:
                    nbest.append(_NbestPrediction(text="", start_logit=null_start_logit, end_logit=null_end_logit))

                # In very rare edge cases we could only
                # have single null pred. We just create a nonce prediction
                # in this case to avoid failure.
                if len(nbest) == 1:
                    nbest.insert(0, _NbestPrediction(text="empty", start_logit=0.0, end_logit=0.0))

            # In very rare edge cases we could have no valid predictions. So we
            # just create a nonce prediction in this case to avoid failure.
            if not nbest:
                nbest.append(_NbestPrediction(text="empty", start_logit=0.0, end_logit=0.0))

            assert len(nbest) >= 1

            total_scores = []
            best_non_null_entry = None
            for entry in nbest:
                total_scores.append(entry.start_logit + entry.end_logit)
                if not best_non_null_entry:
                    if entry.text:
                        best_non_null_entry = entry

            probs = _compute_softmax(total_scores)

            nbest_json = []
            for (i, entry) in enumerate(nbest):
                output = collections.OrderedDict()
                output["text"] = entry.text
                output["probability"] = probs[i]
                output["start_logit"] = (
                    entry.start_logit if isinstance(entry.start_logit, float) else list(entry.start_logit)
                )
                output["end_logit"] = entry.end_logit if isinstance(entry.end_logit, float) else list(entry.end_logit)
                nbest_json.append(output)

            assert len(nbest_json) >= 1

            if not version_2_with_negative:
                all_predictions[example.qas_id] = nbest_json[0]["text"]
            else:
                # predict "" iff the null score -
                # the score of best non-null > threshold
                score_diff = score_null - best_non_null_entry.start_logit - (best_non_null_entry.end_logit)
                scores_diff_json[example.qas_id] = score_diff
                if score_diff > null_score_diff_threshold:
                    all_predictions[example.qas_id] = ""
                else:
                    all_predictions[example.qas_id] = best_non_null_entry.text
            all_nbest_json[example.qas_id] = nbest_json

        return all_predictions, all_nbest_json, scores_diff_json

    def evaluate_predictions(self, all_predictions, no_answer_probs=None, no_answer_probability_threshold=1.0):
        qas_id_to_has_answer = {example.qas_id: bool(example.answers) for example in self.examples}
        has_answer_qids = [qas_id for qas_id, has_answer in qas_id_to_has_answer.items() if has_answer]
        no_answer_qids = [qas_id for qas_id, has_answer in qas_id_to_has_answer.items() if not has_answer]
        if no_answer_probs is None:
            no_answer_probs = {k: 0.0 for k in all_predictions}

        exact, f1 = self.get_raw_scores(all_predictions)

        exact_threshold = apply_no_ans_threshold(
            exact, no_answer_probs, qas_id_to_has_answer, no_answer_probability_threshold
        )
        f1_threshold = apply_no_ans_threshold(
            f1, no_answer_probs, qas_id_to_has_answer, no_answer_probability_threshold
        )

        evaluation = make_eval_dict(exact_threshold, f1_threshold)

        if has_answer_qids:
            has_ans_eval = make_eval_dict(exact_threshold, f1_threshold, qid_list=has_answer_qids)
            merge_eval(evaluation, has_ans_eval, "HasAns")

        if no_answer_qids:
            no_ans_eval = make_eval_dict(exact_threshold, f1_threshold, qid_list=no_answer_qids)
            merge_eval(evaluation, no_ans_eval, "NoAns")

        if no_answer_probs:
            find_all_best_thresh(evaluation, all_predictions, exact, f1, no_answer_probs, qas_id_to_has_answer)

        return evaluation["best_exact"], evaluation["best_f1"]

    def get_raw_scores(self, preds):
        """
        Computes the exact and f1 scores from the examples
        and the model predictions
        """
        exact_scores = {}
        f1_scores = {}

        for example in self.examples:
            qas_id = example.qas_id
            gold_answers = [answer["text"] for answer in example.answers if normalize_answer(answer["text"])]

            if not gold_answers:
                # For unanswerable questions,
                # only correct answer is empty string
                gold_answers = [""]

            if qas_id not in preds:
                logging.warning("Missing prediction for %s" % qas_id)
                continue

            prediction = preds[qas_id]
            exact_scores[qas_id] = max(exact_match_score(a, prediction) for a in gold_answers)
            f1_scores[qas_id] = max(f1_score(a, prediction) for a in gold_answers)

        return exact_scores, f1_scores

    def evaluate(
        self,
        unique_ids,
        start_logits,
        end_logits,
        n_best_size,
        max_answer_length,
        do_lower_case,
        version_2_with_negative,
        null_score_diff_threshold,
    ):

        (all_predictions, all_nbest_json, scores_diff_json) = self.get_predictions(
            unique_ids,
            start_logits,
            end_logits,
            n_best_size,
            max_answer_length,
            do_lower_case,
            version_2_with_negative,
            null_score_diff_threshold,
        )

        exact_match, f1 = self.evaluate_predictions(all_predictions)

        return exact_match, f1, all_predictions, all_nbest_json


class SquadProcessor(DataProcessor):
    """
    Processor for the SQuAD data set.
    used by the version 1.1 and version 2.0 of SQuAD, respectively.
    """

    def __init__(self, data_file, mode):
        self.data_file = data_file
        self.mode = mode

    def __init__(self, data_file, mode):
        self.data_file = data_file
        self.mode = mode

    def get_examples(self):
        if self.data_file is None:
            raise ValueError("SquadProcessor should be instantiated")

        with open(self.data_file, "r", encoding="utf-8") as reader:
            input_data = json.load(reader)["data"]
        return self._create_examples(input_data, set_type=self.mode)

    def _create_examples(self, input_data, set_type):
        examples = []
        for entry in tqdm(input_data):
            title = entry["title"]
            for paragraph in entry["paragraphs"]:
                context_text = paragraph["context"]
                for qa in paragraph["qas"]:
                    qas_id = qa["id"]
                    question_text = qa["question"]
                    start_position_character = None
                    answer_text = None
                    answers = []
                    if "is_impossible" in qa:
                        is_impossible = qa["is_impossible"]
                    else:
                        is_impossible = False

                    if not is_impossible:
                        if set_type == "train" or set_type == "eval":
                            answer = qa["answers"][0]
                            answer_text = answer["text"]
                            start_position_character = answer["answer_start"]
                        if set_type == "eval":
                            answers = qa["answers"]

                    example = SquadExample(
                        qas_id=qas_id,
                        question_text=question_text,
                        context_text=context_text,
                        answer_text=answer_text,
                        start_position_character=start_position_character,
                        title=title,
                        is_impossible=is_impossible,
                        answers=answers,
                    )

                    examples.append(example)
        return examples


class SquadExample(object):
    """
    A single training/test example for the Squad dataset, as loaded from disk.
    Args:
        qas_id: The example's unique identifier
        question_text: The question string
        context_text: The context string
        answer_text: The answer string
        start_position_character: The character position of the start of
            the answer, 0 indexed
        title: The title of the example
        answers: None by default, this is used during evaluation.
            Holds answers as well as their start positions.
        is_impossible: False by default, set to True if the example has
            no possible answer.
    """

    def __init__(
        self,
        qas_id,
        question_text,
        context_text,
        answer_text,
        start_position_character,
        title,
        answers=[],
        is_impossible=False,
    ):
        self.qas_id = qas_id
        self.question_text = question_text
        self.context_text = context_text
        self.answer_text = answer_text
        self.title = title
        self.is_impossible = is_impossible
        self.answers = answers

        self.start_position, self.end_position = 0, 0

        doc_tokens = []
        char_to_word_offset = []
        prev_is_whitespace = True

        # Split on whitespace so that different tokens
        # may be attributed to their original position.
        # ex: context_text = ["hi yo"]
        #     char_to_word_offset = [0, 0, 0, 1, 1]
        #     doc_tokens = ["hi", "yo"]
        for c in self.context_text:
            if is_whitespace(c):
                prev_is_whitespace = True
            else:
                if prev_is_whitespace:
                    doc_tokens.append(c)
                else:
                    doc_tokens[-1] += c
                prev_is_whitespace = False
            char_to_word_offset.append(len(doc_tokens) - 1)

        self.doc_tokens = doc_tokens
        self.char_to_word_offset = char_to_word_offset

        # Start end end positions only has a value during evaluation.
        if start_position_character is not None and not is_impossible:
            # start_position is index of word, end_position inclusive
            self.start_position = char_to_word_offset[start_position_character]
            self.end_position = char_to_word_offset[
                min(start_position_character + len(answer_text) - 1, len(char_to_word_offset) - 1)
            ]



# -*- coding: utf-8 -*-

# =============================================================================
# Copyright (c) 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

# Sadly have to import the whole "nemo" python module to avoid circular dependencies.
# Moreover, at that point nemo module doesn't contain "core", so during "python module registration"
# nothing from nemo.core, including e.g. types (so we cannot use them for "python 3 type hints").
from nemo.utils.neural_graph.object_registry import ObjectRegistry


class NeuralGraphManager(ObjectRegistry):
    def __init__(self):
        """
            Constructor. Initializes the manager. Sets active graph to None.
        """
        super().__init__("graph")
        self._active_graph = None

    def __eq__(self, other):
        """
            Checks if two managers have the same content.
            Args:
                other: A second manager object.
        """
        if not isinstance(other, ObjectRegistry):
            return False
        return super().__eq__(other)

    def summary(self) -> str:
        """
            Returns:
                A summary of the graphs on the list.
        """
        # Line "decorator".
        summary = "\n" + 113 * '=' + "\n"
        summary += "Registry of {}s:\n".format(self._base_type_name)
        for graph in self:
            summary += " * {} ({}) [{}]\n".format(graph.name, len(graph), graph.operation_mode)
        # Line "decorator".
        summary += 113 * '='
        return summary

    @property
    def active_graph(self) -> "NeuralGraph":
        """
            Property returns the active graph. If there is no active graph, creates a new one.

            Returns:
                The active graph object.
        """
        # Create a new graph - training is the default.
        if self._active_graph is None:
            # Import core here (to avoid circular dependency between core-utils).
            from nemo.core import NeuralGraph, OperationMode

            # Create a new "default" graph. Default mode: both.
            new_graph = NeuralGraph(operation_mode=OperationMode.both)
            new_graph._name = self.register(new_graph, None)
            # Set the newly created graph as active.
            self._active_graph = new_graph

        # Return the graph.
        return self._active_graph

    @active_graph.setter
    def active_graph(self, graph: "NeuralGraph"):
        """
            Property sets the active graph.

            Args:
                graph: Neural graph object that will become active.
        """
        # Activate the graph.
        self._active_graph = graph



# Copyright (C) NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The "add_port_docs" decorator is needed to nicely generate neural types in Sphynx for input and output ports

__all__ = [
    'add_port_docs',
]

import functools
import sys

import wrapt


def _normalize_docstring(docstring):
    """Normalizes the docstring.
    Replaces tabs with spaces, removes leading and trailing blanks lines, and
    removes any indentation.
    Copied from PEP-257:
    https://www.python.org/dev/peps/pep-0257/#handling-docstring-indentation
    Args:
        docstring: the docstring to normalize
    Returns:
        The normalized docstring
    """
    if not docstring:
        return ''
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    # (we use sys.maxsize because sys.maxint doesn't exist in Python 3)
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)
    # Return a single string:
    return '\n'.join(trimmed)


def add_port_docs(wrapped=None, instance=None, value=''):
    if wrapped is None:
        return functools.partial(add_port_docs, value=value)

    @wrapt.decorator
    def wrapper(wrapped, instance=None, args=None, kwargs=None):
        return wrapped(*args, **kwargs)

    decorated = wrapper(wrapped)
    try:
        port_2_ntype = decorated(instance)
    except:
        port_2_ntype = None

    port_description = ""
    if port_2_ntype is not None:
        for port, ntype in port_2_ntype.items():
            port_description += "* *" + port + "* : " + str(ntype)
            port_description += "\n\n"

    __doc__ = _normalize_docstring(wrapped.__doc__) + '\n\n' + str(port_description)
    __doc__ = _normalize_docstring(__doc__)

    wrapt.FunctionWrapper.__setattr__(decorated, "__doc__", __doc__)

    return decorated



# ! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""Setup for pip package."""

import codecs
import os
import subprocess
import sys
from distutils import cmd as distutils_cmd
from distutils import log as distutils_log
from itertools import chain

import setuptools


def is_build_action():
    if len(sys.argv) <= 1:
        return False

    BUILD_TOKENS = ["egg_info", "dist", "bdist", "sdist", "install", "build", "develop", "style", "clean"]

    if any([sys.argv[1].startswith(x) for x in BUILD_TOKENS]):
        return True
    else:
        return False


if is_build_action():
    os.environ['NEMO_PACKAGE_BUILDING'] = 'True'

from nemo.package_info import (
    __contact_emails__,
    __contact_names__,
    __description__,
    __download_url__,
    __homepage__,
    __keywords__,
    __license__,
    __package_name__,
    __repository_url__,
    __version__,
)

if os.path.exists('nemo/README.md'):
    with open("nemo/README.md", "r") as fh:
        long_description = fh.read()
    long_description_content_type = "text/markdown"

elif os.path.exists('README.rst'):
    # codec is used for consistent encoding
    long_description = codecs.open(
        os.path.join(os.path.abspath(os.path.dirname(__file__)), 'README.rst'), 'r', 'utf-8',
    ).read()
    long_description_content_type = "text/x-rst"

else:
    long_description = 'See ' + __homepage__


###############################################################################
#                             Dependency Loading                              #
# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% #


def req_file(filename, folder="requirements"):
    with open(os.path.join(folder, filename)) as f:
        content = f.readlines()
    # you may also want to remove whitespace characters
    # Example: `\n` at the end of each line
    return [x.strip() for x in content]


install_requires = req_file("requirements.txt")

extras_require = {
    # User packages
    'docker': req_file("requirements_docker.txt"),
    'test': req_file("requirements_test.txt"),
    # Collections Packages
    'asr': req_file("requirements_asr.txt"),
    'cv': req_file("requirements_cv.txt"),
    'nlp': req_file("requirements_nlp.txt"),
    'simple_gan': req_file("requirements_simple_gan.txt"),
    'tts': req_file("requirements_tts.txt"),
}

extras_require['all'] = list(chain(extras_require.values()))

# TTS depends on ASR and NLP
extras_require['tts'] = list(chain([extras_require['tts'], extras_require['asr'], extras_require['nlp']]))

tests_requirements = extras_require["test"]


###############################################################################
#                            Code style checkers                              #
# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% #


class StyleCommand(distutils_cmd.Command):
    __LINE_WIDTH = 119
    __ISORT_BASE = (
        'isort '
        # These two lines makes isort compatible with black.
        '--multi-line=3 --trailing-comma --force-grid-wrap=0 '
        f'--use-parentheses --line-width={__LINE_WIDTH} -rc -ws'
    )
    __BLACK_BASE = f'black --skip-string-normalization --line-length={__LINE_WIDTH}'
    description = 'Checks overall project code style.'
    user_options = [
        ('scope=', None, 'Folder of file to operate within.'),
        ('fix', None, 'True if tries to fix issues in-place.'),
    ]

    def __call_checker(self, base_command, scope, check):
        command = list(base_command)

        command.append(scope)

        if check:
            command.extend(['--check', '--diff'])

        self.announce(
            msg='Running command: %s' % str(' '.join(command)), level=distutils_log.INFO,
        )

        return_code = subprocess.call(command)

        return return_code

    def _isort(self, scope, check):
        return self.__call_checker(base_command=self.__ISORT_BASE.split(), scope=scope, check=check,)

    def _black(self, scope, check):
        return self.__call_checker(base_command=self.__BLACK_BASE.split(), scope=scope, check=check,)

    def _pass(self):
        self.announce(msg='\033[32mPASS\x1b[0m', level=distutils_log.INFO)

    def _fail(self):
        self.announce(msg='\033[31mFAIL\x1b[0m', level=distutils_log.INFO)

    # noinspection PyAttributeOutsideInit
    def initialize_options(self):
        self.scope = '.'
        self.fix = ''

    def run(self):
        scope, check = self.scope, not self.fix
        isort_return = self._isort(scope=scope, check=check)
        black_return = self._black(scope=scope, check=check)

        if isort_return == 0 and black_return == 0:
            self._pass()
        else:
            self._fail()
            exit(isort_return if isort_return != 0 else black_return)

    def finalize_options(self):
        pass


###############################################################################


setuptools.setup(
    name=__package_name__,
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=__version__,
    description=__description__,
    long_description=long_description,
    long_description_content_type=long_description_content_type,
    # The project's main homepage.
    url=__repository_url__,
    download_url=__download_url__,
    # Author details
    author=__contact_names__,
    author_email=__contact_emails__,
    # maintainer Details
    maintainer=__contact_names__,
    maintainer_email=__contact_emails__,
    # The licence under which the project is released
    license=__license__,
    classifiers=[
        # How mature is this project? Common values are
        #  1 - Planning
        #  2 - Pre-Alpha
        #  3 - Alpha
        #  4 - Beta
        #  5 - Production/Stable
        #  6 - Mature
        #  7 - Inactive
        'Development Status :: 4 - Beta',
        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Information Technology',
        # Indicate what your project relates to
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Image Recognition',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Apache Software License',
        # Supported python versions
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        # Additional Setting
        'Environment :: Console',
        'Natural Language :: English',
        'Operating System :: OS Independent',
    ],
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    setup_requires=['pytest-runner'],
    tests_require=tests_requirements,
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # $ pip install -e ".[all]"
    # $ pip install nemo_toolkit[all]
    extras_require=extras_require,
    # Add in any packaged data.
    include_package_data=True,
    zip_safe=False,
    # PyPI package information.
    keywords=__keywords__,
    # Custom commands.
    cmdclass={'style': StyleCommand},
)



# ! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from unittest import TestCase

import pytest

import nemo.collections.nlp as nemo_nlp
from nemo.collections.nlp.data import SentencePieceTokenizer


class TestSPCTokenizer(TestCase):
    @pytest.mark.unit
    def test_add_special_tokens(self):
        tokenizer = SentencePieceTokenizer("./tests/data/m_common.model")
        special_tokens = nemo_nlp.data.tokenizers.MODEL_SPECIAL_TOKENS['bert']
        tokenizer.add_special_tokens(special_tokens)
        self.assertTrue(tokenizer.vocab_size == tokenizer.original_vocab_size + len(set(special_tokens.values())))

    @pytest.mark.unit
    def test_text_to_tokens(self):
        tokenizer = SentencePieceTokenizer("./tests/data/m_common.model")
        special_tokens = nemo_nlp.data.tokenizers.MODEL_SPECIAL_TOKENS['bert']
        tokenizer.add_special_tokens(special_tokens)

        text = "[CLS] a b c [MASK] e f [SEP] g h i [SEP]"
        tokens = tokenizer.text_to_tokens(text)

        self.assertTrue(len(tokens) == len(text.split()))
        self.assertTrue(tokens.count("[CLS]") == 1)
        self.assertTrue(tokens.count("[MASK]") == 1)
        self.assertTrue(tokens.count("[SEP]") == 2)

    @pytest.mark.unit
    def test_tokens_to_text(self):
        tokenizer = SentencePieceTokenizer("./tests/data/m_common.model")

        text = "[CLS] a b c [MASK] e f [SEP] g h i [SEP]"
        tokens = tokenizer.text_to_tokens(text)
        result = tokenizer.tokens_to_text(tokens)

        self.assertTrue(text == result)

    @pytest.mark.unit
    def test_text_to_ids(self):
        tokenizer = SentencePieceTokenizer("./tests/data/m_common.model")
        special_tokens = nemo_nlp.data.tokenizers.MODEL_SPECIAL_TOKENS['bert']
        tokenizer.add_special_tokens(special_tokens)

        text = "[CLS] a b c [MASK] e f [SEP] g h i [SEP]"
        ids = tokenizer.text_to_ids(text)

        self.assertTrue(len(ids) == len(text.split()))
        self.assertTrue(ids.count(tokenizer.token_to_id("[CLS]")) == 1)
        self.assertTrue(ids.count(tokenizer.token_to_id("[MASK]")) == 1)
        self.assertTrue(ids.count(tokenizer.token_to_id("[SEP]")) == 2)

    @pytest.mark.unit
    def test_ids_to_text(self):
        tokenizer = SentencePieceTokenizer("./tests/data/m_common.model")
        special_tokens = nemo_nlp.data.tokenizers.MODEL_SPECIAL_TOKENS['bert']
        tokenizer.add_special_tokens(special_tokens)

        text = "[CLS] a b c [MASK] e f [SEP] g h i [SEP]"
        ids = tokenizer.text_to_ids(text)
        result = tokenizer.ids_to_text(ids)

        self.assertTrue(text == result)

    @pytest.mark.unit
    def test_tokens_to_ids(self):
        tokenizer = SentencePieceTokenizer("./tests/data/m_common.model")
        special_tokens = nemo_nlp.data.tokenizers.MODEL_SPECIAL_TOKENS['bert']
        tokenizer.add_special_tokens(special_tokens)

        text = "[CLS] a b c [MASK] e f [SEP] g h i [SEP]"
        tokens = tokenizer.text_to_tokens(text)
        ids = tokenizer.tokens_to_ids(tokens)

        self.assertTrue(len(ids) == len(tokens))
        self.assertTrue(ids.count(tokenizer.token_to_id("[CLS]")) == 1)
        self.assertTrue(ids.count(tokenizer.token_to_id("[MASK]")) == 1)
        self.assertTrue(ids.count(tokenizer.token_to_id("[SEP]")) == 2)

    @pytest.mark.unit
    def test_ids_to_tokens(self):
        tokenizer = SentencePieceTokenizer("./tests/data/m_common.model")
        special_tokens = nemo_nlp.data.tokenizers.MODEL_SPECIAL_TOKENS['bert']
        tokenizer.add_special_tokens(special_tokens)

        text = "[CLS] a b c [MASK] e f [SEP] g h i [SEP]"
        tokens = tokenizer.text_to_tokens(text)
        ids = tokenizer.tokens_to_ids(tokens)
        result = tokenizer.ids_to_tokens(ids)

        self.assertTrue(len(result) == len(tokens))

        for i in range(len(result)):
            self.assertTrue(result[i] == tokens[i])



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2018 The Google AI Language Team Authors and The HugginFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================


import math

from nemo.backends.pytorch.nm import TrainableNM
from nemo.collections.nlp.nm.trainables.common.transformer.transformer_decoders import TransformerDecoder
from nemo.collections.nlp.nm.trainables.common.transformer.transformer_encoders import TransformerEncoder
from nemo.collections.nlp.nm.trainables.common.transformer.transformer_generators import (
    BeamSearchSequenceGenerator,
    GreedySequenceGenerator,
)
from nemo.collections.nlp.nm.trainables.common.transformer.transformer_modules import TransformerEmbedding
from nemo.collections.nlp.utils.transformer_utils import transformer_weights_init
from nemo.core.neural_types import ChannelType, LabelsType, MaskType, NeuralType
from nemo.utils.decorators import add_port_docs

__all__ = ['TransformerEncoderNM', 'TransformerDecoderNM', 'GreedyLanguageGeneratorNM', 'BeamSearchTranslatorNM']


class TransformerEncoderNM(TrainableNM):
    """
    Neural module which consists of embedding layer followed by Transformer
    encoder.

    Args:
        vocab_size: size of the vocabulary (number of tokens)
        d_model: hidden size (d_model) of the Transformer
        max_seq_length: maximum allowed length of input sequences, feeding
            longer sequences will cause an error
        embedding_dropout: dropout ratio applied to embeddings
        learn_positional_encodings: bool, whether to learn positional encoding
            or use fixed sinusoidal encodings
        num_layers: number of layers in Transformer encoder
        mask_future: bool, whether to apply triangular future masking to the
            sequence of hidden states (which allows to use it for LM)
        num_attn_heads: number of attention heads
        d_inner: number of neurons in the intermediate part of feed-forward
            network (FFN)
        ffn_dropout: dropout ratio applied to FFN
        attn_score_dropout: dropout ratio applied to attention scores
        attn_layer_dropout: dropout ratio applied to the output of attn layer
        hidden_act: activation function applied in intermediate FFN module
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        input_ids: ids of input tokens
        input_mask_src: input mask
        """
        return {
            "input_ids": NeuralType(('B', 'T'), ChannelType()),
            "input_mask_src": NeuralType(('B', 'T'), ChannelType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        hidden_states: outputs hidden states
        """
        return {"hidden_states": NeuralType(('B', 'T', 'D'), ChannelType())}

    def __init__(
        self,
        vocab_size,
        d_model,
        d_inner,
        max_seq_length,
        num_layers,
        num_attn_heads,
        ffn_dropout=0.0,
        embedding_dropout=0.0,
        attn_score_dropout=0.0,
        attn_layer_dropout=0.0,
        learn_positional_encodings=False,
        hidden_act='relu',
        mask_future=False,
    ):
        super().__init__()

        self.embedding_layer = TransformerEmbedding(
            vocab_size=vocab_size,
            hidden_size=d_model,
            max_sequence_length=max_seq_length,
            embedding_dropout=embedding_dropout,
            learn_positional_encodings=learn_positional_encodings,
        )
        self.encoder = TransformerEncoder(
            num_layers=num_layers,
            hidden_size=d_model,
            mask_future=mask_future,
            num_attention_heads=num_attn_heads,
            inner_size=d_inner,
            ffn_dropout=ffn_dropout,
            hidden_act=hidden_act,
            attn_score_dropout=attn_score_dropout,
            attn_layer_dropout=attn_layer_dropout,
        )

        std_init_range = 1 / math.sqrt(d_model)
        self.apply(lambda module: transformer_weights_init(module, std_init_range))
        self.to(self._device)

    def forward(self, input_ids, input_mask_src):
        hidden_states = self.embedding_layer(input_ids)
        hidden_states = self.encoder(hidden_states, input_mask_src)
        return hidden_states


class TransformerDecoderNM(TrainableNM):
    """
    Neural module which consists of embedding layer followed by Transformer
    decoder.

    Args:
        vocab_size: size of the vocabulary (number of tokens)
        d_model: hidden size (d_model) of the Transformer
        max_seq_length: maximum allowed length of input sequences, feeding
            longer sequences will cause an error
        embedding_dropout: dropout ratio applied to embeddings
        learn_positional_encodings: bool, whether to learn positional encoding
            or use fixed sinusoidal encodings
        num_layers: number of layers in Transformer decoder
        num_attn_heads: number of attention heads
        d_inner: number of neurons in the intermediate part of feed-forward
            network (FFN)
        ffn_dropout: dropout ratio applied to FFN
        attn_score_dropout: dropout ratio applied to attention scores
        attn_layer_dropout: dropout ratio applied to the output of attn layer
        hidden_act: activation function applied in intermediate FFN module
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        input_ids_tgt: ids of target sequence
        hidden_states_src: input hidden states 
        input_mask_src: input token mask
        input_mask_tgt: target token mask
        """
        return {
            "input_ids_tgt": NeuralType(('B', 'T'), LabelsType()),
            "hidden_states_src": NeuralType(('B', 'T', 'D'), ChannelType()),
            "input_mask_src": NeuralType(('B', 'T'), ChannelType()),
            "input_mask_tgt": NeuralType(('B', 'T'), MaskType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        hidden_states: output hidden states
        """
        return {"hidden_states": NeuralType(('B', 'T', 'D'), ChannelType())}

    def __init__(
        self,
        vocab_size,
        d_model,
        d_inner,
        num_layers,
        max_seq_length,
        num_attn_heads,
        ffn_dropout=0.0,
        embedding_dropout=0.0,
        attn_score_dropout=0.0,
        attn_layer_dropout=0.0,
        learn_positional_encodings=False,
        hidden_act='relu',
        use_full_attention=True,
    ):
        super().__init__()

        self.embedding_layer = TransformerEmbedding(
            vocab_size=vocab_size,
            hidden_size=d_model,
            max_sequence_length=max_seq_length,
            embedding_dropout=embedding_dropout,
            learn_positional_encodings=learn_positional_encodings,
        )
        self.decoder = TransformerDecoder(
            num_layers=num_layers,
            hidden_size=d_model,
            num_attention_heads=num_attn_heads,
            inner_size=d_inner,
            ffn_dropout=ffn_dropout,
            hidden_act=hidden_act,
            attn_score_dropout=attn_score_dropout,
            attn_layer_dropout=attn_layer_dropout,
            use_full_attention=use_full_attention,
        )

        std_init_range = 1 / math.sqrt(d_model)
        self.apply(lambda module: transformer_weights_init(module, std_init_range))
        self.to(self._device)

    def forward(self, input_ids_tgt, hidden_states_src, input_mask_src, input_mask_tgt):
        hidden_states_tgt = self.embedding_layer(input_ids_tgt)
        hidden_states = self.decoder(hidden_states_tgt, input_mask_tgt, hidden_states_src, input_mask_src)
        return hidden_states


class GreedyLanguageGeneratorNM(TrainableNM):
    """
    Neural module for greedy text generation with language model

    Args:
        decoder: module which maps input_ids into hidden_states
        log_softmax: module which maps hidden_states into log_probs
        max_seq_length: maximum allowed length of generated sequences
        pad_token: index of padding token in the vocabulary
        bos_token: index of beginning of sequence token in the vocabulary
        eos_token: index of end of sequence token in the vocabulary
        batch_size: size of the batch of generated sequences if no starting
            tokens are provided
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        input_ids:  input ids
        """
        return {"input_ids": NeuralType(('B', 'T'), ChannelType())}

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        output ids: output ids
        """
        return {"output_ids": NeuralType(('B', 'T'), ChannelType())}

    def __init__(self, decoder, log_softmax, max_seq_length, pad_token, bos_token, eos_token, batch_size=1):
        super().__init__()

        self.generator = GreedySequenceGenerator(
            decoder.embedding_layer,
            decoder.decoder,
            log_softmax,
            max_sequence_length=max_seq_length,
            pad=pad_token,
            bos=bos_token,
            eos=eos_token,
            batch_size=batch_size,
        )

    @property
    def num_weights(self):
        return 0

    def forward(self, input_ids):
        output_ids = self.generator(decoder_input_ids=input_ids)
        return output_ids


class BeamSearchTranslatorNM(TrainableNM):
    """
    Neural module for beam search translation generation

    Args:
        decoder: module which maps input_ids into hidden_states
        log_softmax: module which maps hidden_states into log_probs
        max_seq_length: maximum allowed length of generated sequences
        pad_token: index of padding token in the vocabulary
        bos_token: index of beginning of sequence token in the vocabulary
        eos_token: index of end of sequence token in the vocabulary
        batch_size: size of the batch of generated sequences if no starting
            tokens are provided
        beam_size: size of the beam
        max_delta_length: maximum allowed difference between generated output
            and input sequence in case of conditional decoding
        length_penalty: parameter which penalizes shorter sequences
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        hidden_states_src: input hidden states
        input_mask_src: input mask
        """
        return {
            "hidden_states_src": NeuralType(('B', 'T', 'C'), ChannelType()),
            "input_mask_src": NeuralType(('B', 'T'), ChannelType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        output_ids: output ids
        """
        return {"output_ids": NeuralType(('B', 'T'), ChannelType())}

    @property
    def num_weights(self):
        return 0

    def __init__(
        self,
        decoder,
        log_softmax,
        max_seq_length,
        pad_token,
        bos_token,
        eos_token,
        batch_size=1,
        beam_size=4,
        max_delta_length=50,
        length_penalty=0,
    ):
        super().__init__()

        self.generator = BeamSearchSequenceGenerator(
            decoder.embedding_layer,
            decoder.decoder,
            log_softmax,
            max_sequence_length=max_seq_length,
            max_delta_length=max_delta_length,
            pad=pad_token,
            bos=bos_token,
            eos=eos_token,
            batch_size=batch_size,
            beam_size=beam_size,
            len_pen=length_penalty,
        )

    def forward(self, hidden_states_src, input_mask_src):
        output_ids = self.generator(encoder_hidden_states=hidden_states_src, encoder_input_mask=input_mask_src)
        return output_ids



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2019 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""
This file contains code artifacts adapted from the original implementation:
https://github.com/google-research/google-research/blob/master/schema_guided_dst/baseline/data_utils.py
"""

from nemo import logging

__all__ = ['InputExample', 'STR_DONTCARE', 'STATUS_OFF', 'STATUS_ACTIVE', 'STATUS_DONTCARE', 'truncate_seq_pair']

STR_DONTCARE = "dontcare"

# These are used to represent the status of slots (off, active, dontcare) and
# intents (off, active) in dialogue state tracking.
STATUS_OFF = 0
STATUS_ACTIVE = 1
STATUS_DONTCARE = 2
STATUS_CARRY = 3


class InputExample(object):
    """An example for training/inference."""

    def __init__(
        self,
        schema_config,
        service_schema,
        example_id="NONE",
        example_id_num=[],
        is_real_example=False,
        tokenizer=None,
    ):
        """Constructs an InputExample.

        Args:
          max_seq_length: The maximum length of the sequence. Sequences longer than
            this value will be truncated.
          service_schema: A ServiceSchema object wrapping the schema for the service
            corresponding to this example.
          example_id: Unique identifier for the example, like: 'train-1_00000-00-Restaurants_1'
          example_id_num: dialogue_id and turn_id combined and service id combined into a list of ints,
            like: [1, 0, 0, 18]
          is_real_example: Indicates if an example is real or used for padding in a
            minibatch.
          tokenizer (Tokenizer): such as NemoBertTokenizer
        """
        self.schema_config = schema_config
        self.service_schema = service_schema
        self.example_id = example_id
        self.example_id_num = example_id_num
        self._add_carry_value = service_schema._add_carry_value
        self._add_carry_status = service_schema._add_carry_status

        self.is_real_example = is_real_example
        self._max_seq_length = schema_config["MAX_SEQ_LENGTH"]
        self._tokenizer = tokenizer
        if self.is_real_example and self._tokenizer is None:
            raise ValueError("Must specify tokenizer when input is a real example.")

        self.user_utterance = ''
        self.system_utterance = ''
        # The id of each subword in the vocabulary for BERT.
        self.utterance_ids = [0] * self._max_seq_length
        # Denotes the identity of the sequence. Takes values 0 (system utterance) and 1 (user utterance).
        self.utterance_segment = [0] * self._max_seq_length
        # Mask which takes the value 0 for padded tokens and 1 otherwise.
        self.utterance_mask = [0] * self._max_seq_length
        # Start and inclusive end character indices in the original utterance
        # corresponding to the tokens. This is used to obtain the character indices
        # from the predicted subword indices during inference.
        # NOTE: A positive value indicates the character indices in the user
        # utterance whereas a negative value indicates the character indices in the
        # system utterance. The indices are offset by 1 to prevent ambiguity in the
        # 0 index, which could be in either the user or system utterance by the
        # above convention. Now the 0 index corresponds to padded tokens.
        self.start_char_idx = [0] * self._max_seq_length
        self.end_char_idx = [0] * self._max_seq_length

        # Number of categorical slots present in the service.
        self.num_categorical_slots = 0
        # The status of each categorical slot in the service.
        self.categorical_slot_status = [STATUS_OFF] * schema_config["MAX_NUM_CAT_SLOT"]
        # Masks out categorical status for padded cat slots
        self.cat_slot_status_mask = [0] * len(self.categorical_slot_status)
        # Number of values taken by each categorical slot.
        self.num_categorical_slot_values = [0] * schema_config["MAX_NUM_CAT_SLOT"]
        # The index of the correct value for each categorical slot.
        self.categorical_slot_values = [0] * schema_config["MAX_NUM_CAT_SLOT"]
        # Masks out categorical slots values for slots not used in the service
        self.cat_slot_values_mask = [
            [0] * schema_config["MAX_NUM_VALUE_PER_CAT_SLOT"] for _ in range(schema_config["MAX_NUM_CAT_SLOT"])
        ]

        # Number of non-categorical slots present in the service.
        self.num_noncategorical_slots = 0
        # The status of each non-categorical slot in the service.
        self.noncategorical_slot_status = [STATUS_OFF] * schema_config["MAX_NUM_NONCAT_SLOT"]
        # Masks out non-categorical status for padded cat slots
        self.noncat_slot_status_mask = [0] * len(self.noncategorical_slot_status)
        # The index of the starting subword corresponding to the slot span for a
        # non-categorical slot value.
        self.noncategorical_slot_value_start = [0] * schema_config["MAX_NUM_NONCAT_SLOT"]
        # The index of the ending (inclusive) subword corresponding to the slot span
        # for a non-categorical slot value.
        self.noncategorical_slot_value_end = [0] * schema_config["MAX_NUM_NONCAT_SLOT"]

        # Total number of slots present in the service. All slots are included here
        # since every slot can be requested.
        self.num_slots = 0
        # Takes value 1 if the corresponding slot is requested, 0 otherwise.
        self.requested_slot_status = [STATUS_OFF] * (
            schema_config["MAX_NUM_CAT_SLOT"] + schema_config["MAX_NUM_NONCAT_SLOT"]
        )
        # Masks out requested slots that are not used for the service
        self.requested_slot_mask = [0] * len(self.requested_slot_status)

        # Total number of intents present in the service.
        self.num_intents = 0
        # Takes value 1 if the intent is active, 0 otherwise.
        self.intent_status = [STATUS_OFF] * schema_config["MAX_NUM_INTENT"]
        # Masks out intents that are not used for the service, [1] for none intent
        self.intent_status_mask = [1] + [0] * len(self.intent_status)
        # Label for active intent in the turn
        self.intent_status_labels = 0

    @property
    def readable_summary(self):
        """Get a readable dict that summarizes the attributes of an InputExample."""
        seq_length = sum(self.utterance_mask)
        utt_toks = self._tokenizer.ids_to_tokens(self.utterance_ids[:seq_length])
        utt_tok_mask_pairs = list(zip(utt_toks, self.utterance_segment[:seq_length]))
        active_intents = [
            self.service_schema.get_intent_from_id(idx)
            for idx, s in enumerate(self.intent_status)
            if s == STATUS_ACTIVE
        ]
        if len(active_intents) > 1:
            raise ValueError("Should not have multiple active intents in a single service.")
        active_intent = active_intents[0] if active_intents else ""
        slot_values_in_state = {}
        for idx, s in enumerate(self.categorical_slot_status):
            if s == STATUS_ACTIVE:
                value_id = self.categorical_slot_values[idx]
                slot_values_in_state[
                    self.service_schema.get_categorical_slot_from_id(idx)
                ] = self.service_schema.get_categorical_slot_value_from_id(idx, value_id)
            elif s == STATUS_DONTCARE:
                slot_values_in_state[self.service_schema.get_categorical_slot_from_id(idx)] = STR_DONTCARE
        for idx, s in enumerate(self.noncategorical_slot_status):
            if s == STATUS_ACTIVE:
                slot = self.service_schema.get_non_categorical_slot_from_id(idx)
                start_id = self.noncategorical_slot_value_start[idx]
                end_id = self.noncategorical_slot_value_end[idx]
                # Token list is consisted of the subwords that may start with "##". We
                # remove "##" to reconstruct the original value. Note that it's not a
                # strict restoration of the original string. It's primarily used for
                # debugging.
                # ex. ["san", "j", "##ose"] --> "san jose"
                readable_value = " ".join(utt_toks[start_id : end_id + 1]).replace(" ##", "")
                slot_values_in_state[slot] = readable_value
            elif s == STATUS_DONTCARE:
                slot = self.service_schema.get_non_categorical_slot_from_id(idx)
                slot_values_in_state[slot] = STR_DONTCARE

        summary_dict = {
            "utt_tok_mask_pairs": utt_tok_mask_pairs,
            "utt_len": seq_length,
            "num_categorical_slots": self.num_categorical_slots,
            "num_categorical_slot_values": self.num_categorical_slot_values,
            "num_noncategorical_slots": self.num_noncategorical_slots,
            "service_name": self.service_schema.service_name,
            "active_intent": active_intent,
            "slot_values_in_state": slot_values_in_state,
        }
        return summary_dict

    def add_utterance_features(
        self, system_tokens, system_inv_alignments, user_tokens, user_inv_alignments, system_utterance, user_utterance
    ):
        """Add utterance related features input to bert.

        Note: this method modifies the system tokens and user_tokens in place to
        make their total length <= the maximum input length for BERT model.

        Args:
          system_tokens: a list of strings which represents system utterance.
          system_inv_alignments: a list of tuples which denotes the start and end
            charater of the tpken that a bert token originates from in the original
            system utterance.
          user_tokens: a list of strings which represents user utterance.
          user_inv_alignments: a list of tuples which denotes the start and end
            charater of the token that a bert token originates from in the original
            user utterance.
        """
        # Make user-system utterance input (in BERT format)
        # Input sequence length for utterance BERT encoder
        max_utt_len = self._max_seq_length

        # Modify lengths of sys & usr utterance so that length of total utt
        # (including cls_token, setp_token, sep_token) is no more than max_utt_len
        is_too_long = truncate_seq_pair(system_tokens, user_tokens, max_utt_len - 3)
        if is_too_long:
            logging.debug(f'Utterance sequence truncated in example id - {self.example_id}.')

        # Construct the tokens, segment mask and valid token mask which will be
        # input to BERT, using the tokens for system utterance (sequence A) and
        # user utterance (sequence B).
        utt_subword = []
        utt_seg = []
        utt_mask = []
        start_char_idx = []
        end_char_idx = []

        utt_subword.append(self._tokenizer.cls_token)
        utt_seg.append(0)
        utt_mask.append(1)
        start_char_idx.append(0)
        end_char_idx.append(0)

        for subword_idx, subword in enumerate(system_tokens):
            utt_subword.append(subword)
            utt_seg.append(0)
            utt_mask.append(1)
            st, en = system_inv_alignments[subword_idx]
            start_char_idx.append(-(st + 1))
            end_char_idx.append(-(en + 1))

        utt_subword.append(self._tokenizer.sep_token)
        utt_seg.append(0)
        utt_mask.append(1)
        start_char_idx.append(0)
        end_char_idx.append(0)

        for subword_idx, subword in enumerate(user_tokens):
            utt_subword.append(subword)
            utt_seg.append(1)
            utt_mask.append(1)
            st, en = user_inv_alignments[subword_idx]
            start_char_idx.append(st + 1)
            end_char_idx.append(en + 1)

        utt_subword.append(self._tokenizer.sep_token)
        utt_seg.append(1)
        utt_mask.append(1)
        start_char_idx.append(0)
        end_char_idx.append(0)

        utterance_ids = self._tokenizer.tokens_to_ids(utt_subword)

        # Zero-pad up to the BERT input sequence length.
        while len(utterance_ids) < max_utt_len:
            utterance_ids.append(0)
            utt_seg.append(0)
            utt_mask.append(0)
            start_char_idx.append(0)
            end_char_idx.append(0)
        self.utterance_ids = utterance_ids
        self.utterance_segment = utt_seg
        self.utterance_mask = utt_mask
        self.start_char_idx = start_char_idx
        self.end_char_idx = end_char_idx

        self.user_utterances = user_utterance
        self.system_utterance = system_utterance

    def make_copy_with_utterance_features(self):
        """Make a copy of the current example with utterance features."""
        new_example = InputExample(
            schema_config=self.schema_config,
            service_schema=self.service_schema,
            example_id=self.example_id,
            example_id_num=self.example_id_num,
            is_real_example=self.is_real_example,
            tokenizer=self._tokenizer,
        )
        new_example.utterance_ids = list(self.utterance_ids)
        new_example.utterance_segment = list(self.utterance_segment)
        new_example.utterance_mask = list(self.utterance_mask)
        new_example.start_char_idx = list(self.start_char_idx)
        new_example.end_char_idx = list(self.end_char_idx)
        new_example.user_utterance = self.user_utterance
        new_example.system_utterance = self.system_utterance
        return new_example

    def add_categorical_slots(self, state_update, agg_sys_state):
        """Add features for categorical slots."""
        categorical_slots = self.service_schema.categorical_slots
        self.num_categorical_slots = len(categorical_slots)
        for slot_idx, slot in enumerate(categorical_slots):
            values = state_update.get(slot, [])
            # Add categorical slot value features.
            slot_values = self.service_schema.get_categorical_slot_values(slot)
            self.num_categorical_slot_values[slot_idx] = len(slot_values)
            # set slot mask to 1, i.e. the slot exists in the service
            self.cat_slot_status_mask[slot_idx] = 1
            # set the number of active slot values for this slots in the service
            for slot_value_idx in range(len(self.service_schema._categorical_slot_values[slot])):
                self.cat_slot_values_mask[slot_idx][slot_value_idx] = 1

            if not values:
                self.categorical_slot_status[slot_idx] = STATUS_OFF
            elif values[0] == STR_DONTCARE:
                self.categorical_slot_status[slot_idx] = STATUS_DONTCARE
            else:
                value_id = self.service_schema.get_categorical_slot_value_id(slot, values[0])
                if value_id < 0:
                    logging.warning(
                        f"Categorical value not found: EXAMPLE_ID:{self.example_id}, EXAMPLE_ID_NUM:{self.example_id_num}"
                    )
                    logging.warning(f"SYSTEM: {self.system_utterance} || USER: {self.user_utterance}")
                else:
                    if values[0] not in agg_sys_state.get(slot, []):
                        self.categorical_slot_status[slot_idx] = STATUS_ACTIVE
                        self.categorical_slot_values[slot_idx] = value_id
                    else:
                        if self._add_carry_status:
                            self.categorical_slot_status[slot_idx] = STATUS_CARRY
                        else:
                            self.categorical_slot_status[slot_idx] = STATUS_ACTIVE

                        if self._add_carry_value:
                            self.categorical_slot_values[slot_idx] = self.service_schema.get_categorical_slot_value_id(
                                slot, "#CARRYVALUE#"
                            )
                            logging.debug(
                                f"Found slot:{slot}, value:{values[0]}, slot_id:{self.categorical_slot_values[slot_idx]} in prev states: {agg_sys_state}"
                            )
                        else:
                            self.categorical_slot_values[slot_idx] = value_id

    def add_noncategorical_slots(self, state_update, system_span_boundaries, user_span_boundaries):
        """Add features for non-categorical slots."""
        noncategorical_slots = self.service_schema.non_categorical_slots
        self.num_noncategorical_slots = len(noncategorical_slots)
        for slot_idx, slot in enumerate(noncategorical_slots):
            values = state_update.get(slot, [])
            self.noncat_slot_status_mask[slot_idx] = 1
            if not values:
                self.noncategorical_slot_status[slot_idx] = STATUS_OFF
            elif values[0] == STR_DONTCARE:
                self.noncategorical_slot_status[slot_idx] = STATUS_DONTCARE
            else:
                self.noncategorical_slot_status[slot_idx] = STATUS_ACTIVE
                # Add indices of the start and end tokens for the first encountered
                # value. Spans in user utterance are prioritized over the system
                # utterance. If a span is not found, the slot value is ignored.
                if slot in user_span_boundaries:
                    start, end = user_span_boundaries[slot]
                elif slot in system_span_boundaries:
                    start, end = system_span_boundaries[slot]
                else:
                    # A span may not be found because the value was cropped out or because
                    # the value was mentioned earlier in the dialogue. Since this model
                    # only makes use of the last two utterances to predict state updates,
                    # it will fail in such cases.
                    if self._add_carry_status:
                        self.noncategorical_slot_status[slot_idx] = STATUS_CARRY

                    logging.debug(
                        f'"Slot values {str(values)} not found in user or system utterance in example with id - {self.example_id}.'
                    )
                    start = 0
                    end = 0
                self.noncategorical_slot_value_start[slot_idx] = start
                self.noncategorical_slot_value_end[slot_idx] = end

    def add_requested_slots(self, frame):
        all_slots = self.service_schema.slots
        self.num_slots = len(all_slots)
        for slot_idx, slot in enumerate(all_slots):
            self.requested_slot_mask[slot_idx] = 1
            if slot in frame["state"]["requested_slots"]:
                self.requested_slot_status[slot_idx] = STATUS_ACTIVE

    def add_intents(self, frame):
        all_intents = self.service_schema.intents
        self.num_intents = len(all_intents)
        for intent_idx, intent in enumerate(all_intents):
            if intent == frame["state"]["active_intent"]:
                self.intent_status[intent_idx] = STATUS_ACTIVE
                # adding +1 to take none intent into account
                # supports only 1 active intent in the turn
                self.intent_status_labels = intent_idx + 1
            self.intent_status_mask[intent_idx + 1] = 1


# Modified from run_classifier._truncate_seq_pair in the public bert model repo.
# https://github.com/google-research/bert/blob/master/run_classifier.py.
def truncate_seq_pair(tokens_a, tokens_b, max_length):
    """Truncate a seq pair in place so that their total length <= max_length."""
    is_too_long = False
    # This is a simple heuristic which will always truncate the longer sequence
    # one token at a time. This makes more sense than truncating an equal percent
    # of tokens from each, since if one sequence is very short then each token
    # that's truncated likely contains more information than a longer sequence.
    while True:
        total_length = len(tokens_a) + len(tokens_b)
        if total_length <= max_length:
            break
        is_too_long = True
        if len(tokens_a) > len(tokens_b):
            tokens_a.pop()
        else:
            tokens_b.pop()
    return is_too_long



# Copyright (c) 2019 NVIDIA Corporation
import os
import subprocess
import sys
import time
from shutil import copyfile

from nemo.utils import logging
from nemo.utils.decorators import deprecated


# from nemo.utils import logging
@deprecated(
    version=0.11,
    explanation=(
        "Please use nemo.logging instead by using from nemo.utils import logging and logging.info(), "
        "logging.warning() , etc."
    ),
)
def get_logger(unused):
    return logging


# class ContextFilter(logging.Filter):
#     """
#     This is a filter which injects contextual information into the log.
#     Use it when we want to inject worker number into the log message.

#     Usage:
#     logger = get_logger(name)
#     tmp = logging.Formatter(
#         'WORKER %(local_rank)s: %(asctime)s - %(levelname)s - %(message)s')
#     logger.addFilter(ContextFilter(self.local_rank))

#     """

#     def __init__(self, local_rank):
#         super().__init__()
#         self.local_rank = local_rank

#     def filter(self, record):
#         record.local_rank = self.local_rank
#         return True


class ExpManager:
    """ Note: Users should not have to call ExpManager as it is done
    automically inside NeuralFactory. Not all defaults match NeuralFactory
    defaults.
    ExpManager helps create a work directory used to log experiment
    files. It additionally creates a checkpoint directory, tensboard
    directory, tensorboardX.SummaryWriter object,
    copies any files passed with files_to_copy to the work
    directory, and creates loggers used to print to screen and file.

    Args:
    work_dir (str): Directory that Expmanager should either create or
        save log files and directories to.
        Defaults to None and does not create a work directory.
    local_rank (int): None for single-gpu, else the local id for distributed
        setups.
        Defaults to None
    global_rank (int): None for single-gpu, else the global id for distributed
        setups.
        Defaults to None
    use_tb (bool): Whether to create a tensorboardX.SummaryWriter object
        Defaults to True.
    exist_ok (bool): If False, ExpManager will crash if the work_dir
        already exists. Must be false for distributed runs.
    ckpt_dir (str). Whether to create a subdir called ckpt_dir.
        Defaults to which creates a subdir called "checkpoints".
    tb_dir (str): Whether to create a subdir called tb_dir.
        Defaults to which creates a subdir called "tensorboard".
    files_to_copy (list): Should be a list of paths of files that you
        want to copy to work_dir. Useful for copying configs and model
        scrtips.
        Defaults to None which copies no files.
    add_time (bool): Whether to add a datetime ending to work_dir
        Defaults to True.
    broadcast_func (func): Only required if add_time is True and
        distributed. broadcast_func should accept a string that contains
        the datetime suffix such that all ranks are consistent on work_dir
        name.
    """

    def __init__(
        self,
        work_dir=None,
        local_rank=None,
        global_rank=None,
        use_tb=True,
        exist_ok=True,
        ckpt_dir=None,
        tb_dir=None,
        files_to_copy=None,
        add_time=True,
        broadcast_func=None,
    ):
        self.local_rank = local_rank if local_rank is not None else 0
        self.global_rank = global_rank if global_rank is not None else 0
        self.log_file = None
        self.tb_writer = None
        self.work_dir = None
        # Note that if ckpt_dir is None, the default behaviour is to put it
        # under {self.work_dir}/checkpoints
        # If both ckpt_dir and work_dir are None, no ckpt_dir is created
        self.ckpt_dir = None
        # tb_dir behaves the same as ckpt_dir except default folder name is
        # tensorboard instead of checkpoints
        self.tb_dir = tb_dir
        tm_suf = time.strftime('%Y-%m-%d_%H-%M-%S')
        if global_rank is not None and add_time:
            if broadcast_func is None:
                raise ValueError(
                    "local rank was not None, but ExpManager was not passed a "
                    "broadcast function to broadcast the datetime suffix"
                )
            if global_rank == 0:
                broadcast_func(string=tm_suf)
            else:
                tm_suf = broadcast_func(str_len=len(tm_suf))

        # Create work_dir if specified
        if work_dir:
            self.work_dir = work_dir
            # only create tm_sur dir if checkpoints dir is not present in the work_dir
            if add_time:
                self.work_dir = os.path.join(work_dir, tm_suf)
            self.make_dir(self.work_dir, exist_ok)
            self.ckpt_dir = os.path.join(self.work_dir, 'checkpoints')

            if use_tb:
                self.get_tb_writer(exist_ok=exist_ok)

            if files_to_copy and self.global_rank == 0:
                for file in files_to_copy:
                    basename = os.path.basename(file)
                    basename, ending = os.path.splitext(basename)
                    basename = basename + f"_{tm_suf}" + ending
                    copyfile(file, os.path.join(self.work_dir, basename))
            if self.global_rank == 0:
                # Create files for cmd args and git info
                with open(os.path.join(self.work_dir, f'cmd-args_{tm_suf}.log'), 'w') as f:
                    f.write(" ".join(sys.argv))

                # Try to get git hash
                git_repo, git_hash = get_git_hash()
                if git_repo:
                    git_log_file = os.path.join(self.work_dir, f'git-info_{tm_suf}.log')
                    with open(git_log_file, 'w') as f:
                        f.write(f'commit hash: {git_hash}')
                        f.write(get_git_diff())

        # Create loggers
        if bool(work_dir):
            self.add_file_handler_to_logger()
        if use_tb and not work_dir:
            raise ValueError("ExpManager received use_tb as True but did not receive a work_dir")

        if ckpt_dir:
            self.ckpt_dir = ckpt_dir
        if self.ckpt_dir:
            self.make_dir(self.ckpt_dir, exist_ok)

    def add_file_handler_to_logger(self):
        self.log_file = f'{self.work_dir}/log_globalrank-{self.global_rank}_' f'localrank-{self.local_rank}.txt'
        logging.add_file_handler(self.log_file)

    def make_dir(self, dir_, exist_ok):
        # We might want to limit folder creation to only global_rank 0
        os.makedirs(dir_, exist_ok=exist_ok)

    def get_tb_writer(self, tb_dir=None, exist_ok=True):
        if self.global_rank == 0:
            if tb_dir is None:
                if not hasattr(self, 'tb_dir') or self.tb_dir is None:
                    self.tb_dir = os.path.join(self.work_dir, 'tensorboard')
            else:  # if user passes in tb_dir then override the exp's tb_dir
                self.tb_dir = tb_dir
            self.make_dir(self.tb_dir, exist_ok)

            try:
                from torch.utils.tensorboard import SummaryWriter

                self.tb_writer = SummaryWriter(self.tb_dir)
            except ImportError:
                self.tb_writer = None
                logging.info('Not using TensorBoard.')
                logging.info('Install tensorboardX to use TensorBoard')
        return self.tb_writer

    def log_exp_info(self, params, print_everywhere=False):
        if print_everywhere or self.global_rank == 0:
            logging.info("NEMO MODEL'S PARAMETERS")
            for key in params:
                logging.info(f'{key}\t{params[key]}')
            logging.info(f'Experiment output is stored in {self.work_dir}')


def get_git_hash():
    try:
        return (
            True,
            subprocess.check_output(['git', 'rev-parse', 'HEAD'], stderr=subprocess.STDOUT).decode(),
        )
    except subprocess.CalledProcessError as e:
        return False, "{}\n".format(e.output.decode("utf-8"))


def get_git_diff():
    try:
        return subprocess.check_output(['git', 'diff'], stderr=subprocess.STDOUT).decode()
    except subprocess.CalledProcessError as e:
        return "{}\n".format(e.output.decode("utf-8"))
